import jax
import jax.numpy as jnp
from jax import random, grad, jit
import numpy as np
from typing import Optional, Tuple, Dict
from functools import partial


class AttentionMechanism:
    """
    A comprehensive class implementing various attention mechanisms
    for both univariate time series and tabular data using JAX.

    Supported attention types:
    - Scaled Dot-Product Attention
    - Additive (Bahdanau) Attention
    - Multi-Head Attention
    - Self-Attention
    - Temporal Attention (for sequences)
    - Feature Attention (for tabular data)
    - Cross-Attention
    - Context Vector Attention
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dim: int = 64,
        num_heads: int = 4,
        dropout: float = 0.1,
        seed: int = 42,
    ):
        """
        Args:
            input_dim: Dimension of input features
            hidden_dim: Hidden dimension for attention computations
            num_heads: Number of attention heads for multi-head attention
            dropout: Dropout rate
            seed: Random seed for parameter initialization
        """
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.num_heads = num_heads
        self.dropout = dropout

        # Initialize random key
        self.rng = random.PRNGKey(seed)

        # Initialize parameters
        self.params = self._initialize_parameters()

        assert (
            hidden_dim % num_heads == 0
        ), "hidden_dim must be divisible by num_heads"
        self.head_dim = hidden_dim // num_heads

    def _initialize_parameters(self) -> Dict:
        """Initialize all network parameters using JAX"""
        keys = random.split(self.rng, 20)

        def init_weight(key, shape):
            return random.normal(key, shape) * np.sqrt(2.0 / shape[0])

        def init_bias(shape):
            return jnp.zeros(shape)

        params = {
            # Scaled Dot-Product Attention
            "query_w": init_weight(keys[0], (self.input_dim, self.hidden_dim)),
            "query_b": init_bias((self.hidden_dim,)),
            "key_w": init_weight(keys[1], (self.input_dim, self.hidden_dim)),
            "key_b": init_bias((self.hidden_dim,)),
            "value_w": init_weight(keys[2], (self.input_dim, self.hidden_dim)),
            "value_b": init_bias((self.hidden_dim,)),
            # Additive Attention
            "additive_query_w": init_weight(
                keys[3], (self.input_dim, self.hidden_dim)
            ),
            "additive_query_b": init_bias((self.hidden_dim,)),
            "additive_key_w": init_weight(
                keys[4], (self.input_dim, self.hidden_dim)
            ),
            "additive_key_b": init_bias((self.hidden_dim,)),
            "additive_v_w": init_weight(keys[5], (self.hidden_dim, 1)),
            "additive_v_b": init_bias((1,)),
            # Multi-Head Attention
            "mha_query_w": init_weight(
                keys[6], (self.input_dim, self.hidden_dim)
            ),
            "mha_query_b": init_bias((self.hidden_dim,)),
            "mha_key_w": init_weight(
                keys[7], (self.input_dim, self.hidden_dim)
            ),
            "mha_key_b": init_bias((self.hidden_dim,)),
            "mha_value_w": init_weight(
                keys[8], (self.input_dim, self.hidden_dim)
            ),
            "mha_value_b": init_bias((self.hidden_dim,)),
            "mha_output_w": init_weight(
                keys[9], (self.hidden_dim, self.hidden_dim)
            ),
            "mha_output_b": init_bias((self.hidden_dim,)),
            # Feature Attention
            "feature_w1": init_weight(
                keys[10], (self.input_dim, self.hidden_dim)
            ),
            "feature_b1": init_bias((self.hidden_dim,)),
            "feature_w2": init_weight(
                keys[11], (self.hidden_dim, self.input_dim)
            ),
            "feature_b2": init_bias((self.input_dim,)),
            # Temporal Attention
            "temporal_query_w": init_weight(
                keys[12], (self.input_dim, self.hidden_dim)
            ),
            "temporal_query_b": init_bias((self.hidden_dim,)),
            "temporal_key_w": init_weight(
                keys[13], (self.input_dim, self.hidden_dim)
            ),
            "temporal_key_b": init_bias((self.hidden_dim,)),
            # Context Vector Attention
            "context_vector": random.normal(keys[14], (1, 1, self.hidden_dim)),
            "context_query_w": init_weight(
                keys[15], (self.hidden_dim, self.hidden_dim)
            ),
            "context_query_b": init_bias((self.hidden_dim,)),
            "context_key_w": init_weight(
                keys[16], (self.input_dim, self.hidden_dim)
            ),
            "context_key_b": init_bias((self.hidden_dim,)),
            "context_value_w": init_weight(
                keys[17], (self.input_dim, self.hidden_dim)
            ),
            "context_value_b": init_bias((self.hidden_dim,)),
        }

        return params

    @staticmethod
    @jit
    def _apply_dropout(
        x: jnp.ndarray,
        key: jax.random.PRNGKey,
        rate: float,
        training: bool = True,
    ) -> jnp.ndarray:
        """Apply dropout"""
        if training and rate > 0:
            keep_prob = 1 - rate
            mask = random.bernoulli(key, keep_prob, x.shape)
            return jnp.where(mask, x / keep_prob, 0)
        return x

    @partial(jit, static_argnums=(0,))
    def scaled_dot_product_attention(
        self,
        query: jnp.ndarray,
        key: jnp.ndarray,
        value: jnp.ndarray,
        params: Dict,
        mask: Optional[jnp.ndarray] = None,
        training: bool = False,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Scaled Dot-Product Attention

        Args:
            query: (batch_size, seq_len, input_dim) or (batch_size, input_dim)
            key: (batch_size, seq_len, input_dim)
            value: (batch_size, seq_len, input_dim)
            params: Parameter dictionary
            mask: Optional mask (batch_size, seq_len)
            training: Whether in training mode

        Returns:
            context: Attended context vector
            attention_weights: Attention weights
        """
        # Project inputs
        Q = jnp.dot(query, params["query_w"]) + params["query_b"]
        K = jnp.dot(key, params["key_w"]) + params["key_b"]
        V = jnp.dot(value, params["value_w"]) + params["value_b"]

        # Compute attention scores
        scores = jnp.matmul(Q, jnp.swapaxes(K, -2, -1))
        scores = scores / jnp.sqrt(self.hidden_dim)

        # Apply mask if provided
        if mask is not None:
            scores = jnp.where(mask == 0, -1e9, scores)

        # Compute attention weights
        attention_weights = jax.nn.softmax(scores, axis=-1)

        # Apply attention to values
        context = jnp.matmul(attention_weights, V)

        return context, attention_weights

    @partial(jit, static_argnums=(0,))
    def additive_attention(
        self,
        query: jnp.ndarray,
        key: jnp.ndarray,
        value: jnp.ndarray,
        params: Dict,
        mask: Optional[jnp.ndarray] = None,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Additive (Bahdanau) Attention

        Args:
            query: (batch_size, hidden_dim) or (batch_size, 1, hidden_dim)
            key: (batch_size, seq_len, hidden_dim)
            value: (batch_size, seq_len, hidden_dim)
            params: Parameter dictionary
            mask: Optional mask

        Returns:
            context: Attended context vector
            attention_weights: Attention weights
        """
        # Ensure query has seq_len dimension
        if query.ndim == 2:
            query = jnp.expand_dims(query, axis=1)

        # Project query and key
        Q = (
            jnp.dot(query, params["additive_query_w"])
            + params["additive_query_b"]
        )
        K = jnp.dot(key, params["additive_key_w"]) + params["additive_key_b"]

        # Additive attention: score = v^T tanh(W_q Q + W_k K)
        combined = jnp.tanh(Q + K)
        scores = (
            jnp.dot(combined, params["additive_v_w"]) + params["additive_v_b"]
        )
        scores = jnp.squeeze(scores, axis=-1)

        # Apply mask if provided
        if mask is not None:
            scores = jnp.where(mask == 0, -1e9, scores)

        # Compute attention weights
        attention_weights = jax.nn.softmax(scores, axis=-1)

        # Apply attention to values
        context = jnp.matmul(jnp.expand_dims(attention_weights, axis=1), value)
        context = jnp.squeeze(context, axis=1)

        return context, attention_weights

    @partial(jit, static_argnums=(0,))
    def multi_head_attention(
        self,
        query: jnp.ndarray,
        key: jnp.ndarray,
        value: jnp.ndarray,
        params: Dict,
        mask: Optional[jnp.ndarray] = None,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Multi-Head Attention

        Args:
            query: (batch_size, seq_len_q, input_dim)
            key: (batch_size, seq_len_k, input_dim)
            value: (batch_size, seq_len_v, input_dim)
            params: Parameter dictionary
            mask: Optional mask

        Returns:
            output: Multi-head attention output
            attention_weights: Attention weights from all heads
        """
        batch_size = query.shape[0]

        # Project and reshape for multi-head attention
        Q = jnp.dot(query, params["mha_query_w"]) + params["mha_query_b"]
        K = jnp.dot(key, params["mha_key_w"]) + params["mha_key_b"]
        V = jnp.dot(value, params["mha_value_w"]) + params["mha_value_b"]

        Q = Q.reshape(batch_size, -1, self.num_heads, self.head_dim)
        K = K.reshape(batch_size, -1, self.num_heads, self.head_dim)
        V = V.reshape(batch_size, -1, self.num_heads, self.head_dim)

        # Transpose for attention: (batch, num_heads, seq_len, head_dim)
        Q = jnp.transpose(Q, (0, 2, 1, 3))
        K = jnp.transpose(K, (0, 2, 1, 3))
        V = jnp.transpose(V, (0, 2, 1, 3))

        # Compute attention scores
        scores = jnp.matmul(Q, jnp.swapaxes(K, -2, -1)) / jnp.sqrt(
            self.head_dim
        )

        # Apply mask if provided
        if mask is not None:
            mask_expanded = jnp.expand_dims(jnp.expand_dims(mask, 1), 2)
            scores = jnp.where(mask_expanded == 0, -1e9, scores)

        # Attention weights
        attention_weights = jax.nn.softmax(scores, axis=-1)

        # Apply attention to values
        context = jnp.matmul(attention_weights, V)

        # Reshape back: (batch, seq_len, hidden_dim)
        context = jnp.transpose(context, (0, 2, 1, 3))
        context = context.reshape(batch_size, -1, self.hidden_dim)

        # Final linear projection
        output = (
            jnp.dot(context, params["mha_output_w"]) + params["mha_output_b"]
        )

        return output, attention_weights

    @partial(jit, static_argnums=(0,))
    def self_attention(
        self, x: jnp.ndarray, params: Dict, mask: Optional[jnp.ndarray] = None
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """Self-Attention mechanism"""
        return self.scaled_dot_product_attention(x, x, x, params, mask)

    @partial(jit, static_argnums=(0,))
    def temporal_attention(
        self, x: jnp.ndarray, params: Dict, mask: Optional[jnp.ndarray] = None
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Temporal Attention for time series data

        Args:
            x: (batch_size, seq_len, input_dim)
            params: Parameter dictionary
            mask: Optional mask

        Returns:
            context: Temporally attended context
            attention_weights: Temporal attention weights
        """
        # Use last time step as query
        query = x[:, -1:, :]

        Q = (
            jnp.dot(query, params["temporal_query_w"])
            + params["temporal_query_b"]
        )
        K = jnp.dot(x, params["temporal_key_w"]) + params["temporal_key_b"]

        # Compute attention scores
        scores = jnp.matmul(Q, jnp.swapaxes(K, -2, -1)) / jnp.sqrt(
            self.hidden_dim
        )
        scores = jnp.squeeze(scores, axis=1)

        # Apply mask if provided
        if mask is not None:
            scores = jnp.where(mask == 0, -1e9, scores)

        # Attention weights
        attention_weights = jax.nn.softmax(scores, axis=-1)

        # Apply attention
        context = jnp.matmul(jnp.expand_dims(attention_weights, axis=1), x)
        context = jnp.squeeze(context, axis=1)

        return context, attention_weights

    @partial(jit, static_argnums=(0,))
    def feature_attention_tabular(
        self, x: jnp.ndarray, params: Dict
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Feature Attention for tabular data

        Args:
            x: (batch_size, num_features)
            params: Parameter dictionary

        Returns:
            output: Feature-weighted output
            attention_weights: Feature importance weights
        """
        # Compute feature attention weights
        hidden = jnp.dot(x, params["feature_w1"]) + params["feature_b1"]
        hidden = jnp.tanh(hidden)
        logits = jnp.dot(hidden, params["feature_w2"]) + params["feature_b2"]
        attention_weights = jax.nn.softmax(logits, axis=-1)

        # Apply attention to features
        output = x * attention_weights

        return output, attention_weights

    @partial(jit, static_argnums=(0,))
    def context_vector_attention(
        self, x: jnp.ndarray, params: Dict, mask: Optional[jnp.ndarray] = None
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Context Vector Attention
        Uses a learnable global context vector as the query.

        Args:
            x: (batch_size, seq_len, input_dim)
            params: Parameter dictionary
            mask: Optional mask (batch_size, seq_len)

        Returns:
            context: Global context representation (batch_size, hidden_dim)
            attention_weights: Attention weights (batch_size, seq_len)
        """
        batch_size = x.shape[0]

        # Expand context vector for batch
        context_vec = jnp.broadcast_to(
            params["context_vector"], (batch_size, 1, self.hidden_dim)
        )

        # Project context vector and input
        Q = (
            jnp.dot(context_vec, params["context_query_w"])
            + params["context_query_b"]
        )
        K = jnp.dot(x, params["context_key_w"]) + params["context_key_b"]
        V = jnp.dot(x, params["context_value_w"]) + params["context_value_b"]

        # Compute attention scores
        scores = jnp.matmul(Q, jnp.swapaxes(K, -2, -1)) / jnp.sqrt(
            self.hidden_dim
        )
        scores = jnp.squeeze(scores, axis=1)

        # Apply mask if provided
        if mask is not None:
            scores = jnp.where(mask == 0, -1e9, scores)

        # Compute attention weights
        attention_weights = jax.nn.softmax(scores, axis=-1)

        # Apply attention to values
        context = jnp.matmul(jnp.expand_dims(attention_weights, axis=1), V)
        context = jnp.squeeze(context, axis=1)

        return context, attention_weights

    @partial(jit, static_argnums=(0,))
    def cross_attention(
        self,
        query: jnp.ndarray,
        key_value: jnp.ndarray,
        params: Dict,
        mask: Optional[jnp.ndarray] = None,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """Cross-Attention between two different sequences"""
        return self.scaled_dot_product_attention(
            query, key_value, key_value, params, mask
        )

    def __call__(
        self,
        x: jnp.ndarray,
        attention_type: str = "scaled_dot_product",
        query: Optional[jnp.ndarray] = None,
        key_value: Optional[jnp.ndarray] = None,
        mask: Optional[jnp.ndarray] = None,
        training: bool = False,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Forward pass with specified attention mechanism

        Args:
            x: Input tensor
            attention_type: Type of attention to use
            query: Optional query for cross-attention
            key_value: Optional key-value for cross-attention
            mask: Optional mask
            training: Whether in training mode

        Returns:
            output: Attention output
            attention_weights: Attention weights
        """
        if attention_type == "scaled_dot_product":
            return self.scaled_dot_product_attention(
                x, x, x, self.params, mask, training
            )
        elif attention_type == "additive":
            return self.additive_attention(
                x[:, -1:, :], x, x, self.params, mask
            )
        elif attention_type == "multi_head":
            return self.multi_head_attention(x, x, x, self.params, mask)
        elif attention_type == "self":
            return self.self_attention(x, self.params, mask)
        elif attention_type == "temporal":
            return self.temporal_attention(x, self.params, mask)
        elif attention_type == "feature":
            return self.feature_attention_tabular(x, self.params)
        elif attention_type == "cross":
            if query is None or key_value is None:
                raise ValueError(
                    "Cross-attention requires both query and key_value"
                )
            return self.cross_attention(query, key_value, self.params, mask)
        elif attention_type == "context_vector":
            return self.context_vector_attention(x, self.params, mask)
        else:
            raise ValueError(f"Unknown attention type: {attention_type}")



import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted
from sklearn.kernel_approximation import RBFSampler


def _get_estimator_type(estimator):
    """Determine if estimator is classifier or regressor."""
    if hasattr(estimator, "_estimator_type"):
        return estimator._estimator_type
    elif hasattr(estimator, "predict_proba") or hasattr(
        estimator, "decision_function"
    ):
        return "classifier"
    elif hasattr(estimator, "predict"):
        return "regressor"
    else:
        return None


class RandomFourierEstimator(BaseEstimator):
    def __init__(
        self, estimator, n_components=100, gamma=1.0, random_state=None
    ):
        """
        Random Fourier Features transformation with a given estimator.

        Parameters:
        - estimator: A scikit-learn estimator (classifier, regressor, etc.).
        - n_components: Number of random Fourier features.
        - gamma: Hyperparameter for RBF kernel approximation.
        - random_state: Random state for reproducibility.
        """
        self.estimator = estimator
        self.n_components = n_components
        self.gamma = gamma
        self.random_state = random_state

        # Dynamically set the estimator type and appropriate mixin
        estimator_type = _get_estimator_type(estimator)
        if estimator_type == "classifier":
            self._estimator_type = "classifier"
            # Add ClassifierMixin to the class hierarchy
            if not isinstance(self, ClassifierMixin):
                self.__class__ = type(
                    self.__class__.__name__,
                    (self.__class__, ClassifierMixin),
                    dict(self.__class__.__dict__),
                )
        elif estimator_type == "regressor":
            self._estimator_type = "regressor"
            # Add RegressorMixin to the class hierarchy
            if not isinstance(self, RegressorMixin):
                self.__class__ = type(
                    self.__class__.__name__,
                    (self.__class__, RegressorMixin),
                    dict(self.__class__.__dict__),
                )

    def fit(self, X, y=None):
        """
        Fit the Random Fourier feature transformer and the estimator.
        """
        X = check_array(X)

        # Initialize and fit the Random Fourier Feature transformer
        self.rff_ = RBFSampler(
            n_components=self.n_components,
            gamma=self.gamma,
            random_state=self.random_state,
        )
        X_transformed = self.rff_.fit_transform(X)

        # Fit the underlying estimator on the transformed data
        self.estimator.fit(X_transformed, y)

        return self

    def partial_fit(self, X, y, classes=None):
        """
        Incrementally fit the Random Fourier feature transformer and the estimator.
        """
        X = check_array(X)

        # Check if RFF transformer is already fitted
        if not hasattr(self, "rff_"):
            # First call - fit the transformer
            self.rff_ = RBFSampler(
                n_components=self.n_components,
                gamma=self.gamma,
                random_state=self.random_state,
            )
            X_transformed = self.rff_.fit_transform(X)
        else:
            # Subsequent calls - only transform
            X_transformed = self.rff_.transform(X)

        # If estimator supports partial_fit, we use it, otherwise raise an error
        if hasattr(self.estimator, "partial_fit"):
            self.estimator.partial_fit(X_transformed, y, classes=classes)
        else:
            raise ValueError(
                f"The estimator {type(self.estimator).__name__} does not support partial_fit method."
            )

        return self

    def predict(self, X):
        """
        Predict using the Random Fourier transformed data.
        """
        check_is_fitted(self, ["rff_"])
        X = check_array(X)

        # Transform the input data
        X_transformed = self.rff_.transform(X)

        # Predict using the underlying estimator
        return self.estimator.predict(X_transformed)

    def predict_proba(self, X):
        """
        Predict class probabilities (only for classifiers).
        """
        if (
            not hasattr(self, "_estimator_type")
            or self._estimator_type != "classifier"
        ):
            raise AttributeError(
                "predict_proba is not available for this estimator type."
            )

        check_is_fitted(self, ["rff_"])
        X = check_array(X)

        if not hasattr(self.estimator, "predict_proba"):
            raise ValueError(
                f"The estimator {type(self.estimator).__name__} does not support predict_proba."
            )

        # Transform the input data
        X_transformed = self.rff_.transform(X)

        # Predict probabilities using the underlying estimator
        return self.estimator.predict_proba(X_transformed)

    def predict_log_proba(self, X):
        """
        Predict class log probabilities (only for classifiers).
        """
        if (
            not hasattr(self, "_estimator_type")
            or self._estimator_type != "classifier"
        ):
            raise AttributeError(
                "predict_log_proba is not available for this estimator type."
            )

        check_is_fitted(self, ["rff_"])
        X = check_array(X)

        if not hasattr(self.estimator, "predict_log_proba"):
            raise ValueError(
                f"The estimator {type(self.estimator).__name__} does not support predict_log_proba."
            )

        # Transform the input data
        X_transformed = self.rff_.transform(X)

        return self.estimator.predict_log_proba(X_transformed)

    def decision_function(self, X):
        """
        Decision function (only for classifiers).
        """
        if (
            not hasattr(self, "_estimator_type")
            or self._estimator_type != "classifier"
        ):
            raise AttributeError(
                "decision_function is not available for this estimator type."
            )

        check_is_fitted(self, ["rff_"])
        X = check_array(X)

        if not hasattr(self.estimator, "decision_function"):
            raise ValueError(
                f"The estimator {type(self.estimator).__name__} does not support decision_function."
            )

        # Transform the input data
        X_transformed = self.rff_.transform(X)

        return self.estimator.decision_function(X_transformed)

    def score(self, X, y):
        """
        Evaluate the model performance.
        """
        check_is_fitted(self, ["rff_"])
        X = check_array(X)

        # Transform the input data
        X_transformed = self.rff_.transform(X)

        # Evaluate using the underlying estimator's score method
        return self.estimator.score(X_transformed, y)

    @property
    def classes_(self):
        """Classes labels (only for classifiers)."""
        if (
            hasattr(self, "_estimator_type")
            and self._estimator_type == "classifier"
        ):
            return getattr(self.estimator, "classes_", None)
        else:
            raise AttributeError(
                "classes_ is not available for this estimator type."
            )

    def get_params(self, deep=True):
        """
        Get parameters for this estimator.
        """
        params = {}

        # Get estimator parameters with proper prefixing
        if deep:
            estimator_params = self.estimator.get_params(deep=True)
            for key, value in estimator_params.items():
                params[f"estimator__{key}"] = value

        # Add our own parameters
        params.update(
            {
                "estimator": self.estimator,
                "n_components": self.n_components,
                "gamma": self.gamma,
                "random_state": self.random_state,
            }
        )

        return params

    def set_params(self, **params):
        """
        Set the parameters of this estimator.
        """
        # Separate our parameters from estimator parameters
        our_params = {}
        estimator_params = {}

        for param, value in params.items():
            if param.startswith("estimator__"):
                # Remove the 'estimator__' prefix
                estimator_params[param[11:]] = value
            elif param in [
                "estimator",
                "n_components",
                "gamma",
                "random_state",
            ]:
                our_params[param] = value
            else:
                # Assume it's an estimator parameter without prefix
                estimator_params[param] = value

        # Set our parameters
        for param, value in our_params.items():
            setattr(self, param, value)

        # If estimator changed, update the estimator type
        if "estimator" in our_params:
            self.__init__(
                self.estimator, self.n_components, self.gamma, self.random_state
            )

        # Set estimator parameters
        if estimator_params:
            self.estimator.set_params(**estimator_params)

        # If RFF parameters changed and model is fitted, we need to refit
        if hasattr(self, "rff_") and (
            "n_components" in our_params
            or "gamma" in our_params
            or "random_state" in our_params
        ):
            # Remove the fitted transformer so it gets recreated on next fit
            delattr(self, "rff_")

        return self



from .mts import MTS
from .mlarch import MLARCH
from .classical import ClassicalMTS
from .stackedmts import MTSStacker
from .multioutputmts import MultiOutputMTS

__all__ = ["MTS", "MLARCH", "ClassicalMTS",
 "MTSStacker", "MultiOutputMTS"]



import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_X_y, check_array


class ElasticNet2Regressor(BaseEstimator, RegressorMixin):
    def __init__(self, n_hidden_features=100, alpha=1.0, l1_ratio=0.5, 
                 lambd=0.1, activation_name='tanh', a=0.01,
                 max_iter=1000, tol=1e-4, random_state=None):
        self.n_hidden_features = n_hidden_features
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.lambd = lambd
        self.activation_name = activation_name
        self.a = a
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state
    
    def _activation(self, Z):
        if self.activation_name == 'relu':
            return np.maximum(0, Z)
        elif self.activation_name == 'tanh':
            return np.tanh(Z)
        elif self.activation_name == 'sigmoid':
            return 1 / (1 + np.exp(-Z))
        elif self.activation_name == 'prelu':
            return np.where(Z > 0, Z, self.a * Z)
        elif self.activation_name == 'elu':
            return np.where(Z > 0, Z, self.a * (np.exp(Z) - 1))
        else:
            raise ValueError(f"Unknown activation: {self.activation_name}")
    
    def fit(self, X, y):
        X, y = check_X_y(X, y)
        rng = np.random.RandomState(self.random_state)
        
        # Standardize inputs
        self.X_mean_ = X.mean(axis=0)
        self.X_std_ = X.std(axis=0) + 1e-8
        X_scaled = (X - self.X_mean_) / self.X_std_
        
        # Center response
        self.y_mean_ = y.mean()
        y_centered = y - self.y_mean_
        
        # Random feature mapping
        self.W_in_ = rng.randn(X.shape[1], self.n_hidden_features)
        self.b_in_ = rng.randn(self.n_hidden_features)
        H = self._activation(X_scaled @ self.W_in_ + self.b_in_)
        
        # Doubly-constrained optimization with Elastic Net
        beta = np.zeros(self.n_hidden_features)
        
        for _ in range(self.max_iter):
            beta_old = beta.copy()
            
            # Gradient descent step with projection
            grad = H.T @ (H @ beta - y_centered) / len(y)
            step = 0.01 / (1 + self.alpha * (1 - self.l1_ratio))
            
            # Soft thresholding (L1)
            beta = beta - step * grad
            threshold = step * self.alpha * self.l1_ratio
            beta = np.sign(beta) * np.maximum(np.abs(beta) - threshold, 0)
            
            # L2 projection (constraint)
            norm = np.linalg.norm(beta)
            if norm > self.lambd:
                beta = beta * (self.lambd / norm)
            
            if np.linalg.norm(beta - beta_old) < self.tol:
                break
        
        self.beta_ = beta
        return self
    
    def predict(self, X):
        X = check_array(X)
        X_scaled = (X - self.X_mean_) / self.X_std_
        H = self._activation(X_scaled @ self.W_in_ + self.b_in_)
        return H @ self.beta_ + self.y_mean_


import numpy as np
import pandas as pd
from collections import namedtuple
from scipy.stats import norm
from sklearn.linear_model import Ridge, RidgeCV
import matplotlib.pyplot as plt


class MLARCH:
    """Machine Learning-agnostic ARCH for nearly-stationary time series (e.g., returns)

    Parameters
    ----------
    model_mean : object
        Model for mean component
    model_sigma : object
        Model for volatility component (sklearn regressor)
    model_residuals : object
        Model for standardized residuals
    lags_vol : int, default=10
        Number of lags for squared residuals in volatility model
    """

    def __init__(self, model_mean, model_sigma, model_residuals, lags_vol=10):
        self.model_mean = model_mean
        self.model_sigma = model_sigma
        self.model_residuals = model_residuals
        self.lags_vol = lags_vol

    def _create_lags(self, y, lags):
        """Create lagged feature matrix"""
        n = len(y)
        if n <= lags:
            raise ValueError(f"Series length {n} must be > lags {lags}")
        X = np.zeros((n - lags, lags))
        for i in range(lags):
            X[:, i] = y[i: (n - lags + i)]
        return X

    def fit(self, y, **kwargs):
        """Fit the MLARCH model

        Parameters
        ----------
        y : array-like
            Target time series (should be stationary, e.g., returns)

        Returns
        -------
        self
        """
        # Format input
        if isinstance(y, (pd.Series, pd.DataFrame)):
            y = y.values
        y = y.ravel()

        if len(y) < self.lags_vol + 20:
            raise ValueError(f"Need at least {self.lags_vol + 20} observations")

        # Step 1: Fit mean model
        self.model_mean.fit(y.reshape(-1, 1))
        mean_residuals = self.model_mean.residuals_.ravel()

        # Step 2: Fit ARCH volatility model on lagged squared residuals
        resid_squared = mean_residuals**2
        X_vol = self._create_lags(resid_squared, self.lags_vol)
        y_vol = np.log(resid_squared[self.lags_vol:] + 1e-8)

        self.model_sigma.fit(X_vol, y_vol)

        # Get fitted volatility
        fitted_log_sigma = self.model_sigma.predict(X_vol)
        fitted_sigma = np.exp(fitted_log_sigma)

        # Step 3: Compute standardized residuals with proper scaling
        standardized_residuals = mean_residuals[self.lags_vol:] / np.sqrt(
            fitted_sigma
        )

        # Enforce zero mean and unit variance
        self.z_mean_ = np.mean(standardized_residuals)
        self.z_std_ = np.std(standardized_residuals)
        standardized_residuals = (
            standardized_residuals - self.z_mean_
        ) / self.z_std_

        # Step 4: Fit residuals model
        self.model_residuals.fit(standardized_residuals.reshape(-1, 1))

        # Store for prediction
        self.last_residuals_squared_ = resid_squared[-self.lags_vol:]

        # Store diagnostics
        self.fitted_volatility_mean_ = np.mean(np.sqrt(fitted_sigma))
        self.fitted_volatility_std_ = np.std(np.sqrt(fitted_sigma))

        return self

    def predict(self, h=5, level=95, return_sims=False):
        """Predict future values

        Parameters
        ----------
        h : int
            Forecast horizon
        level : int
            Confidence level for prediction intervals
        return_sims : bool
            If True, return full simulation paths

        Returns
        -------
        DescribeResult
            Named tuple with mean, sims, lower, upper
        """
        DescribeResult = namedtuple(
            "DescribeResult", ("mean", "sims", "lower", "upper")
        )

        # Get mean forecast
        mean_forecast = self.model_mean.predict(h=h).values.ravel()

        # Recursive ARCH volatility forecasting
        sigma_forecast = np.zeros(h)
        current_lags = self.last_residuals_squared_.copy()

        for i in range(h):
            X_t = current_lags.reshape(1, -1)
            log_sigma_t = self.model_sigma.predict(X_t)[0]
            sigma_forecast[i] = np.exp(log_sigma_t)
            # Update lags with predicted variance
            current_lags = np.append(current_lags[1:], sigma_forecast[i])

        # Predict standardized residuals and rescale
        z_forecast_normalized = self.model_residuals.predict(h=h).values.ravel()
        z_forecast = z_forecast_normalized * self.z_std_ + self.z_mean_

        # Combine: μ + z × σ
        point_forecast = mean_forecast + z_forecast * np.sqrt(sigma_forecast)

        # Generate prediction intervals
        sims = None
        if return_sims:
            preds_z_for_sims = self.model_residuals.predict(h=h)
            if hasattr(preds_z_for_sims, "sims") and isinstance(
                preds_z_for_sims.sims, pd.DataFrame
            ):
                sims_z_normalized = preds_z_for_sims.sims
                n_sims = sims_z_normalized.shape[1]

                sims = np.zeros((h, n_sims))
                for sim_idx in range(n_sims):
                    # Rescale simulations
                    z_sim = (
                        sims_z_normalized.iloc[:, sim_idx].values * self.z_std_
                        + self.z_mean_
                    )
                    sims[:, sim_idx] = mean_forecast + z_sim * np.sqrt(
                        sigma_forecast
                    )

                alpha = 1 - level / 100
                lower_bound = np.quantile(sims, alpha / 2, axis=1)
                upper_bound = np.quantile(sims, 1 - alpha / 2, axis=1)
            else:
                # Fallback to Gaussian
                z_score = norm.ppf(1 - (1 - level / 100) / 2)
                margin = z_score * np.sqrt(sigma_forecast) * self.z_std_
                lower_bound = point_forecast - margin
                upper_bound = point_forecast + margin
        else:
            # Gaussian intervals with proper scaling
            z_score = norm.ppf(1 - (1 - level / 100) / 2)
            margin = z_score * np.sqrt(sigma_forecast) * self.z_std_
            lower_bound = point_forecast - margin
            upper_bound = point_forecast + margin

        return DescribeResult(point_forecast, sims, lower_bound, upper_bound)



# Authors: Thierry Moudiki
#
# License: BSD 3 Clear

import numpy as np
import platform
import warnings
from sklearn.base import BaseEstimator
from sklearn.base import RegressorMixin
from numpy.linalg import inv
from ..utils import matrixops as mo

from .utils import get_beta

try:
    import jax.numpy as jnp
    from jax import device_put
    from jax.numpy.linalg import inv as jinv
except ImportError:
    pass


class RidgeRegressor(BaseEstimator, RegressorMixin):
    """Ridge.

    Attributes:

        reg_lambda: float
            regularization parameter.

        backend: str
            type of backend; must be in ('cpu', 'gpu', 'tpu')

    """

    def __init__(self, reg_lambda=0.1, backend="cpu"):
        assert backend in (
            "cpu",
            "gpu",
            "tpu",
        ), "`backend` must be in ('cpu', 'gpu', 'tpu')"

        sys_platform = platform.system()

        if (sys_platform == "Windows") and (backend in ("gpu", "tpu")):
            warnings.warn(
                "No GPU/TPU computing on Windows yet, backend set to 'cpu'"
            )
            backend = "cpu"

        self.reg_lambda = reg_lambda
        self.backend = backend
        self.coef_ = None

    def fit(self, X, y, **kwargs):
        """Fit matrixops (classifier) to training data (X, y)

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to self.cook_training_set.

        Returns:

            self: object.

        """
        self.ym, centered_y = mo.center_response(y)
        self.xm = X.mean(axis=0)
        self.xsd = X.std(axis=0)
        self.xsd[self.xsd == 0] = 1  # avoid division by zero
        X_ = (X - self.xm[None, :]) / self.xsd[None, :]

        if self.backend == "cpu":
            if len(centered_y.shape) <= 1:
                eye_term = np.sqrt(self.reg_lambda) * np.eye(X.shape[1])
                X_ = np.row_stack((X_, eye_term))
                y_ = np.concatenate((centered_y, np.zeros(X.shape[1])))
                beta_info = get_beta(X_, y_)
                self.coef_ = beta_info[0]
            else:
                try:
                    eye_term = np.sqrt(self.reg_lambda) * np.eye(X.shape[1])
                    X_ = np.row_stack((X_, eye_term))
                    y_ = np.row_stack(
                        (
                            centered_y,
                            np.zeros((eye_term.shape[0], centered_y.shape[1])),
                        )
                    )
                    beta_info = get_beta(X_, y_)
                    self.coef_ = beta_info[0]
                except Exception:
                    x = inv(
                        mo.crossprod(X_) + self.reg_lambda * np.eye(X_.shape[1])
                    )
                    hat_matrix = mo.tcrossprod(x, X_)
                    self.coef_ = mo.safe_sparse_dot(hat_matrix, centered_y)
            return self

        x = jinv(
            mo.crossprod(X_, backend=self.backend)
            + self.reg_lambda * jnp.eye(X_.shape[1])
        )

        hat_matrix = mo.tcrossprod(x, X_, backend=self.backend)
        self.coef_ = mo.safe_sparse_dot(
            hat_matrix, centered_y, backend=self.backend
        )
        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to `predict_proba`

        Returns:

            model predictions: {array-like}

        """
        X_ = (X - self.xm[None, :]) / self.xsd[None, :]

        if self.backend == "cpu":
            if isinstance(self.ym, float):
                return self.ym + mo.safe_sparse_dot(X_, self.coef_)
            return self.ym[None, :] + mo.safe_sparse_dot(X_, self.coef_)

        # if self.backend in ("gpu", "tpu"):
        if isinstance(self.ym, float):
            return self.ym + mo.safe_sparse_dot(
                X_, self.coef_, backend=self.backend
            )
        return self.ym[None, :] + mo.safe_sparse_dot(
            X_, self.coef_, backend=self.backend
        )



import jax
import jax.numpy as jnp
import numpy as np
from scipy.optimize import minimize_scalar
from typing import Optional, Tuple, Union, Dict
from jax import random, vmap
from functools import partial
from sklearn.base import BaseEstimator, RegressorMixin


class RandomFourierFeaturesRidge(BaseEstimator, RegressorMixin):
    """
    Random Fourier Features with Bayesian Ridge Regression.

    Implements both standard (MLE) and Bayesian versions with uncertainty quantification.
    Uses data augmentation for L2 regularization via jnp.lstsq.
    """

    def __init__(
        self,
        n_features: int = 100,
        gamma: float = 1.0,
        alpha: float = 1e-6,
        include_bias: bool = True,
        random_seed: int = 42,
    ):
        """
        Parameters:
        -----------
        n_features : int
            Number of random Fourier features (D)
        gamma : float
            RBF kernel parameter: k(x,y) = exp(-gamma * ||x-y||²)
        alpha : float
            Prior precision (inverse variance) for Bayesian version
            Equivalent to regularization strength: lambda = alpha / beta
        include_bias : bool
            Whether to include a bias term
        random_seed : int
            Random seed for reproducibility
        """
        self.n_features = n_features
        self.gamma = gamma
        self.alpha = alpha
        self.include_bias = include_bias
        self.key = random.PRNGKey(random_seed)
        self.is_fitted = False

        # Bayesian parameters
        self.beta = None  # Noise precision (will be estimated from data)
        self.w_mean = None  # Posterior mean of weights
        self.w_cov = None  # Posterior covariance of weights
        self.S_N = None  # Posterior precision matrix

    def _compute_random_features(
        self, X: jnp.ndarray, W: jnp.ndarray, b: jnp.ndarray
    ) -> jnp.ndarray:
        """Compute random Fourier features: sqrt(2/D) * cos(XW + b)"""
        projection = jnp.dot(X, W) + b  # Shape: (n_samples, n_features)
        features = jnp.sqrt(2.0 / self.n_features) * jnp.cos(projection)

        if self.include_bias:
            features = jnp.concatenate(
                [jnp.ones((X.shape[0], 1)), features], axis=1
            )

        return features

    def _init_random_weights(
        self, input_dim: int
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """Initialize random weights and biases for RFF"""
        # Sample from Gaussian distribution for RBF kernel
        # Variance = 2 * gamma for RBF kernel
        self.key, subkey = random.split(self.key)
        W = random.normal(
            subkey, shape=(input_dim, self.n_features)
        ) * jnp.sqrt(2.0 * self.gamma)

        self.key, subkey = random.split(self.key)
        b = random.uniform(
            subkey, shape=(1, self.n_features), minval=0, maxval=2 * jnp.pi
        )

        return W, b

    def fit(
        self,
        X: Union[jnp.ndarray, np.ndarray],
        y: Union[jnp.ndarray, np.ndarray],
        method: str = "bayesian",
        noise_variance: Optional[float] = None,
    ) -> "RandomFourierFeaturesRidge":
        """
        Fit the model using either standard or Bayesian ridge regression.

        Parameters:
        -----------
        X : array-like, shape (n_samples, n_features)
            Training data
        y : array-like, shape (n_samples,) or (n_samples, n_targets)
            Target values
        method : str, either "standard" or "bayesian"
            "standard": Maximum likelihood estimation with L2 regularization
            "bayesian": Full Bayesian inference with uncertainty quantification
        noise_variance : float, optional
            If provided, fixes the noise variance instead of estimating it
        """
        # Convert to JAX arrays if needed
        X = jnp.asarray(X)
        y = jnp.asarray(y)

        if len(y.shape) == 1:
            y = y.reshape(-1, 1)

        n_samples, input_dim = X.shape

        # Initialize random Fourier weights
        self.W, self.b = self._init_random_weights(input_dim)

        # Compute random Fourier features
        Phi = self._compute_random_features(X, self.W, self.b)
        n_basis = Phi.shape[1]  # D + 1 if bias included

        # Store feature matrix and target values for Bayesian updates/likelihood computation
        self.Phi_train = Phi
        self.y_train = y  # Store y_train

        if method == "standard":
            # Standard ridge regression using data augmentation for regularization
            self._fit_standard(Phi, y)
        elif method == "bayesian":
            # Bayesian ridge regression
            self._fit_bayesian(Phi, y, noise_variance)
        else:
            raise ValueError("method must be 'standard' or 'bayesian'")

        self.is_fitted = True
        self.method = method
        self.input_dim = input_dim

        return self

    def _fit_standard(self, Phi: jnp.ndarray, y: jnp.ndarray) -> None:
        """Standard ridge regression using lstsq with data augmentation"""
        n_samples, n_basis = Phi.shape

        # Create augmented data for L2 regularization
        # This is equivalent to adding sqrt(alpha) * I to the design matrix
        sqrt_alpha = jnp.sqrt(self.alpha)
        Phi_aug = jnp.vstack([Phi, sqrt_alpha * jnp.eye(n_basis)])
        y_aug = jnp.vstack([y, jnp.zeros((n_basis, y.shape[1]))])

        # Solve using least squares
        # Note: jnp.linalg.lstsq is more stable than explicit normal equations
        weights, residuals, rank, s = jnp.linalg.lstsq(
            Phi_aug, y_aug, rcond=None
        )

        self.w_mean = weights
        self.weights = weights  # For compatibility

        # Estimate noise variance from residuals
        residuals = y - Phi @ weights
        self.beta = 1.0 / jnp.maximum(jnp.var(residuals), 1e-8)

    def _fit_bayesian(
        self,
        Phi: jnp.ndarray,
        y: jnp.ndarray,
        noise_variance: Optional[float] = None,
    ) -> None:
        """Bayesian ridge regression with evidence approximation"""
        n_samples, n_basis = Phi.shape

        # Initialize precision parameters
        if noise_variance is not None:
            self.beta = 1.0 / noise_variance
        else:
            # Initial estimate of beta from data
            self.beta = 1.0 / jnp.maximum(jnp.var(y), 1e-8)

        # Posterior precision matrix: S_N⁻¹ = alpha * I + beta * ΦᵀΦ
        I = jnp.eye(n_basis)
        PhiT_Phi = Phi.T @ Phi

        # Initialize with prior
        S_N_inv = self.alpha * I

        # Evidence approximation to optimize alpha, beta
        for _ in range(10):  # Iterate to converge on alpha, beta
            # Update posterior mean and covariance
            S_N = jnp.linalg.inv(S_N_inv + self.beta * PhiT_Phi)
            self.w_mean = self.beta * S_N @ Phi.T @ y

            # Update gamma (effective number of parameters)
            eigenvalues = jnp.linalg.eigvalsh(PhiT_Phi)
            gamma_val = jnp.sum(eigenvalues / (self.alpha + eigenvalues))

            # Update alpha and beta (MacKay's fixed point updates)
            if self.alpha > 0:
                self.alpha = gamma_val / jnp.sum(self.w_mean**2)

            if noise_variance is None:
                residuals = y - Phi @ self.w_mean
                self.beta = (n_samples - gamma_val) / jnp.sum(residuals**2)

            # Update precision matrix
            S_N_inv = self.alpha * I

        # Store final covariance
        self.S_N = jnp.linalg.inv(self.alpha * I + self.beta * PhiT_Phi)
        self.w_cov = self.S_N

        # Also store for compatibility
        self.weights = self.w_mean

    def transform(self, X: Union[jnp.ndarray, np.ndarray]) -> jnp.ndarray:
        """Transform input data to random Fourier feature space"""
        if not self.is_fitted:
            raise ValueError("Model must be fitted before transforming")

        X = jnp.asarray(X)
        return self._compute_random_features(X, self.W, self.b)

    def predict(
        self,
        X: Union[jnp.ndarray, np.ndarray],
        return_std: bool = False,
        return_cov: bool = False,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray]]:
        """
        Make predictions, optionally with uncertainty quantification.

        Parameters:
        -----------
        X : array-like, shape (n_samples, n_features)
            Input data
        return_std : bool
            If True, return standard deviation of predictive distribution
        return_cov : bool
            If True, return full covariance matrix of predictive distribution

        Returns:
        --------
        y_pred : jnp.ndarray
            Predictive mean
        y_std or y_cov : jnp.ndarray, optional
            Predictive standard deviation or covariance
        """
        if not self.is_fitted:
            raise ValueError("Model must be fitted before prediction")

        X = jnp.asarray(X)
        Phi = self.transform(X)

        # Predictive mean
        y_pred = Phi @ self.w_mean

        if not return_std and not return_cov:
            return y_pred

        if self.method != "bayesian":
            raise ValueError(
                "Uncertainty quantification only available for Bayesian method"
            )

        # Predictive variance
        if return_cov:
            # Full predictive covariance
            # Σ_pred = (1/β) * I + Φ @ S_N @ Φᵀ
            pred_cov = (1.0 / self.beta) * jnp.eye(
                Phi.shape[0]
            ) + Phi @ self.S_N @ Phi.T
            return y_pred, pred_cov
        else:
            # Diagonal of predictive covariance (standard deviations)
            # σ²_pred = (1/β) + diag(Φ @ S_N @ Φᵀ)
            var_diag = (1.0 / self.beta) + jnp.sum(
                (Phi @ self.S_N) * Phi, axis=1
            )
            y_std = jnp.sqrt(jnp.maximum(var_diag, 0.0)).reshape(-1, 1)
            return y_pred, y_std

    def sample_posterior(
        self,
        X: Union[jnp.ndarray, np.ndarray],
        n_samples: int = 1,
        key: Optional[jax.random.PRNGKey] = None,
    ) -> jnp.ndarray:
        """
        Sample from the posterior predictive distribution.

        Parameters:
        -----------
        X : array-like
            Input data
        n_samples : int
            Number of samples to draw
        key : PRNGKey, optional
            Random key for sampling

        Returns:
        --------
        samples : jnp.ndarray, shape (n_samples, n_test_samples)
            Samples from posterior predictive distribution
        """
        if self.method != "bayesian":
            raise ValueError("Sampling only available for Bayesian method")

        if key is None:
            key = self.key

        X = jnp.asarray(X)
        Phi = self.transform(X)
        n_test = Phi.shape[0]

        # Sample weights from posterior
        key, subkey = random.split(key)
        w_samples = random.multivariate_normal(
            subkey, self.w_mean.flatten(), self.S_N, shape=(n_samples,)
        )

        # Generate predictions for each weight sample
        samples = []
        for i in range(n_samples):
            w_sample = w_samples[i].reshape(-1, 1)
            # Add noise variance
            key, subkey1, subkey2 = random.split(key, 3)
            pred_mean = Phi @ w_sample
            noise = random.normal(subkey2, shape=pred_mean.shape) / jnp.sqrt(
                self.beta
            )
            samples.append(pred_mean + noise)

        return jnp.stack(samples, axis=0)

    def log_marginal_likelihood(self) -> float:
        """
        Compute log marginal likelihood (evidence) for Bayesian model.

        Returns:
        --------
        log_evidence : float
            Log marginal likelihood p(y|X,α,β)
        """
        if self.method != "bayesian":
            raise ValueError(
                "Log marginal likelihood only available for Bayesian method"
            )

        n_samples = self.Phi_train.shape[0]
        n_basis = self.Phi_train.shape[1]

        # Log determinant term
        I = jnp.eye(n_basis)
        A = self.alpha * I + self.beta * self.Phi_train.T @ self.Phi_train
        sign, logdet_A = jnp.linalg.slogdet(A)
        logdet_term = 0.5 * (n_basis * jnp.log(self.alpha) - logdet_A)

        # Data fit term
        residuals = self.y_train - self.Phi_train @ self.w_mean
        data_fit_term = -0.5 * self.beta * jnp.sum(residuals**2)

        # Constant term
        const_term = 0.5 * n_samples * jnp.log(self.beta / (2 * jnp.pi))

        return float(logdet_term + data_fit_term + const_term)

    def get_params(self) -> Dict:
        """Get model parameters"""
        return {
            "n_features": self.n_features,
            "gamma": self.gamma,
            "alpha": self.alpha,
            "beta": self.beta if self.beta is not None else None,
            "method": self.method if hasattr(self, "method") else None,
            "input_dim": self.input_dim if hasattr(self, "input_dim") else None,
        }

    def set_params(self, **params) -> "RandomFourierFeaturesRidge":
        """Set model parameters"""
        for key, value in params.items():
            if hasattr(self, key):
                setattr(self, key, value)
        return self


class RandomFourierFeaturesRidgeGCV(RandomFourierFeaturesRidge):
    """
    Extends RandomFourierFeaturesRidge with GCV for automatic
    regularization parameter selection.
    """

    def __init__(
        self,
        n_features: int = 100,
        gamma: float = 1.0,
        alpha: Optional[float] = None,
        include_bias: bool = True,
        random_seed: int = 42,
    ):
        super().__init__(n_features, gamma, alpha, include_bias, random_seed)
        self.alpha_opt = None  # Stores the GCV-optimized alpha
        self.gcv_score = None  # Stores the optimal GCV score

    def _compute_gcv(
        self,
        alpha: float,
        s_sq: jnp.ndarray,
        U: jnp.ndarray,
        y: jnp.ndarray,
        n_samples: int,
    ) -> float:
        """
        Compute GCV score for a given alpha.

        Parameters:
        -----------
        alpha : float
            Regularization parameter
        s_sq : jnp.ndarray
            Squared singular values of design matrix Φ
        U : jnp.ndarray
            Left singular vectors of Φ
        y : jnp.ndarray
            Target values
        n_samples : int
            Number of data points

        Returns:
        --------
        gcv : float
            GCV score for this alpha
        """
        # Degrees of freedom: df(α) = Σ(σ_j²/(σ_j² + α))
        df = jnp.sum(s_sq / (s_sq + alpha))

        # Compute residual sum of squares efficiently using SVD
        # y_pred = U @ (S²/(S² + α)) @ (U.T @ y)
        Uty = U.T @ y
        shrinkage = s_sq / (s_sq + alpha)
        y_pred = U @ (shrinkage * Uty)
        residuals = y - y_pred
        rss = jnp.sum(residuals**2)

        # GCV formula
        denom = (1.0 - df / n_samples) ** 2
        gcv = (rss / n_samples) / denom

        return float(gcv)

    def fit_gcv(
        self,
        X: Union[jnp.ndarray, np.ndarray],
        y: Union[jnp.ndarray, np.ndarray],
        alpha_range: Tuple[float, float] = (1e-8, 1e4),
        n_alphas: int = 50,
        method: str = "standard",
        optimize: bool = True,
    ) -> "RandomFourierFeaturesRidgeGCV":
        """
        Fit model with GCV-optimized regularization parameter.

        Parameters:
        -----------
        X : array-like
            Training data
        y : array-like
            Target values
        alpha_range : tuple
            (min_alpha, max_alpha) range to search
        n_alphas : int
            Number of alpha values to try in initial grid search
        method : str
            "standard" or "bayesian"
        optimize : bool
            If True, perform fine optimization after grid search

        Returns:
        --------
        self : fitted model
        """
        # Convert to JAX arrays
        X = jnp.asarray(X)
        y = jnp.asarray(y)

        if len(y.shape) == 1:
            y = y.reshape(-1, 1)

        n_samples, input_dim = X.shape

        # Initialize random Fourier weights
        self.W, self.b = self._init_random_weights(input_dim)

        # Compute random Fourier features
        Phi = self._compute_random_features(X, self.W, self.b)

        # Compute SVD of design matrix for efficient GCV computation
        # Φ = U @ diag(S) @ V.T
        U, S, Vt = jnp.linalg.svd(Phi, full_matrices=False)
        s_sq = S**2  # Squared singular values

        # Grid search on log scale
        alphas_grid = jnp.logspace(
            jnp.log10(alpha_range[0]), jnp.log10(alpha_range[1]), n_alphas
        )

        gcv_scores = []
        for alpha in alphas_grid:
            score = self._compute_gcv(float(alpha), s_sq, U, y, n_samples)
            gcv_scores.append(score)

        # Find best alpha from grid
        best_idx = jnp.argmin(jnp.array(gcv_scores))
        alpha_grid_opt = float(alphas_grid[best_idx])

        # Fine optimization using Brent's method
        if optimize:
            # Define objective for scipy optimizer
            def gcv_objective(log_alpha):
                alpha = 10**log_alpha
                return self._compute_gcv(alpha, s_sq, U, y, n_samples)

            # Optimize in log space
            result = minimize_scalar(
                gcv_objective,
                bounds=(jnp.log10(alpha_range[0]), jnp.log10(alpha_range[1])),
                method="bounded",
                options={"xatol": 0.1},  # Tolerance in log10 space
            )

            if result.success:
                alpha_opt = 10**result.x
                gcv_opt = result.fun
            else:
                alpha_opt = alpha_grid_opt
                gcv_opt = gcv_scores[best_idx]
        else:
            alpha_opt = alpha_grid_opt
            gcv_opt = gcv_scores[best_idx]

        # Store optimized parameters
        self.alpha_opt = alpha_opt
        self.gcv_score = gcv_opt
        self.alpha = alpha_opt  # Set as the model's alpha

        # Fit final model with optimized alpha
        if method == "standard":
            self._fit_standard(Phi, y)
        elif method == "bayesian":
            # For Bayesian version, we can use alpha as prior precision
            # Optionally optimize beta too
            self._fit_bayesian(Phi, y)
        else:
            raise ValueError("method must be 'standard' or 'bayesian'")

        self.is_fitted = True
        self.method = method
        self.input_dim = input_dim

        return self

    def fit_gcv_with_path(
        self,
        X: Union[jnp.ndarray, np.ndarray],
        y: Union[jnp.ndarray, np.ndarray],
        alpha_range: Tuple[float, float] = (1e-8, 1e4),
        n_alphas: int = 100,
        method: str = "standard",
    ) -> dict:
        """
        Fit with GCV and return full regularization path.

        Returns:
        --------
        path_info : dict
            Dictionary with alpha values, GCV scores, and metrics
        """
        X = jnp.asarray(X)
        y = jnp.asarray(y)

        if len(y.shape) == 1:
            y = y.reshape(-1, 1)

        n_samples, input_dim = X.shape

        # Initialize random features
        self.W, self.b = self._init_random_weights(input_dim)
        Phi = self._compute_random_features(X, self.W, self.b)

        # Compute SVD
        U, S, Vt = jnp.linalg.svd(Phi, full_matrices=False)
        s_sq = S**2

        # Compute GCV path
        alphas = jnp.logspace(
            jnp.log10(alpha_range[0]), jnp.log10(alpha_range[1]), n_alphas
        )

        gcv_scores = []
        train_errors = []
        effective_dof = []

        for alpha in alphas:
            alpha_val = float(alpha)

            # GCV score
            gcv = self._compute_gcv(alpha_val, s_sq, U, y, n_samples)
            gcv_scores.append(gcv)

            # Effective degrees of freedom
            df = float(jnp.sum(s_sq / (s_sq + alpha_val)))
            effective_dof.append(df)

            # Training error for this alpha
            # Compute weights: w = V @ (S/(S² + α)) @ (U.T @ y)
            Uty = U.T @ y
            shrinkage = S / (s_sq + alpha_val)
            w_alpha = Vt.T @ (shrinkage.reshape(-1, 1) * Uty)
            y_pred = Phi @ w_alpha
            train_err = float(jnp.mean((y - y_pred) ** 2))
            train_errors.append(train_err)

        # Find optimal alpha
        best_idx = jnp.argmin(jnp.array(gcv_scores))
        alpha_opt = float(alphas[best_idx])

        # Fit final model with optimal alpha
        self.alpha = alpha_opt
        if method == "standard":
            self._fit_standard(Phi, y)
        elif method == "bayesian":
            self._fit_bayesian(Phi, y)

        self.is_fitted = True
        self.method = method
        self.input_dim = input_dim
        self.alpha_opt = alpha_opt
        self.gcv_score = gcv_scores[best_idx]

        # Return full path information
        path_info = {
            "alphas": np.array(alphas),
            "gcv_scores": np.array(gcv_scores),
            "train_errors": np.array(train_errors),
            "effective_dof": np.array(effective_dof),
            "alpha_opt": alpha_opt,
            "gcv_opt": gcv_scores[best_idx],
            "dof_opt": effective_dof[best_idx],
        }

        return path_info

    def plot_gcv_path(self, path_info: dict, save_path: str = None):
        """
        Plot GCV regularization path.
        """
        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(2, 2, figsize=(12, 10))

        # Plot 1: GCV score vs alpha
        ax = axes[0, 0]
        ax.semilogx(
            path_info["alphas"], path_info["gcv_scores"], "b-", linewidth=2
        )
        ax.axvline(
            path_info["alpha_opt"],
            color="r",
            linestyle="--",
            label=f'Optimal α = {path_info["alpha_opt"]:.2e}',
        )
        ax.set_xlabel("Regularization α")
        ax.set_ylabel("GCV Score")
        ax.set_title("GCV Score vs Regularization")
        ax.legend()
        ax.grid(True, alpha=0.3)

        # Plot 2: Training error vs alpha
        ax = axes[0, 1]
        ax.loglog(
            path_info["alphas"], path_info["train_errors"], "g-", linewidth=2
        )
        ax.axvline(path_info["alpha_opt"], color="r", linestyle="--")
        ax.set_xlabel("Regularization α")
        ax.set_ylabel("Training MSE")
        ax.set_title("Training Error vs Regularization")
        ax.grid(True, alpha=0.3)

        # Plot 3: Effective DOF vs alpha
        ax = axes[1, 0]
        ax.semilogx(
            path_info["alphas"], path_info["effective_dof"], "m-", linewidth=2
        )
        ax.axvline(path_info["alpha_opt"], color="r", linestyle="--")
        ax.axhline(
            path_info["dof_opt"],
            color="r",
            linestyle=":",
            label=f'DOF at optimum = {path_info["dof_opt"]:.1f}',
        )
        ax.set_xlabel("Regularization α")
        ax.set_ylabel("Effective Degrees of Freedom")
        ax.set_title("Model Complexity vs Regularization")
        ax.legend()
        ax.grid(True, alpha=0.3)

        # Plot 4: GCV vs DOF
        ax = axes[1, 1]
        ax.plot(
            path_info["effective_dof"],
            path_info["gcv_scores"],
            "k-",
            linewidth=2,
        )
        ax.axvline(path_info["dof_opt"], color="r", linestyle="--")
        ax.set_xlabel("Effective Degrees of Freedom")
        ax.set_ylabel("GCV Score")
        ax.set_title("GCV vs Model Complexity")
        ax.grid(True, alpha=0.3)

        plt.suptitle(
            "GCV Regularization Path Analysis", fontsize=14, fontweight="bold"
        )
        plt.tight_layout()

        if save_path:
            plt.savefig(save_path, dpi=150, bbox_inches="tight")

        plt.show()



from .custom import Custom
from .customClassifier import CustomClassifier
from .customRegressor import CustomRegressor
from .customBackpropRegressor import CustomBackPropRegressor

__all__ = [
    "Custom",
    "CustomClassifier",
    "CustomRegressor",
    "CustomBackPropRegressor",
]



from .ridge2Classifier import Ridge2Classifier
from .ridge2Regressor import Ridge2Regressor
from .ridge2MultitaskClassifier import Ridge2MultitaskClassifier
from .ridge2MTSJAX import Ridge2Forecaster
from .ridge2multioutputregressor import Ridge2MultiOutputRegressor

__all__ = [
    "Ridge2Classifier",
    "Ridge2Regressor",
    "Ridge2MultitaskClassifier",
    "Ridge2Forecaster",
    "Ridge2MultiOutputRegressor",
]



# Authors: Thierry Moudiki
#
# License: BSD 3 Clear

import copy
import jax
import numpy as np
import pandas as pd
import platform
import warnings
import sklearn.metrics as skm

from typing import Optional
from collections import namedtuple
from functools import partial
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.model_selection import cross_val_score, train_test_split
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    f1_score,
    roc_auc_score,
    r2_score,
    mean_squared_error,
    mean_absolute_error,
    mean_absolute_percentage_error,
)
from ..utils import activations as ac
from ..utils import matrixops as mo
from ..utils import misc as mx
from ..simulation import (
    generate_sobol,
    generate_uniform,
    generate_hammersley,
    generate_halton,
)
from ..sampling import SubSampler


try:
    import jax.nn as jnn
    import jax.numpy as jnp
except ImportError:
    pass


class Base(BaseEstimator):
    """Base model from which all the other classes inherit.

    This class contains the most important data preprocessing/feature engineering methods.

    Parameters:

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for hidden layer nodes: 'sobol', 'hammersley', 'halton',
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or
            not (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out
            of the training

        direct_link: boolean
            indicates if the original features are included (True) in model's
            fitting or not (False)

        n_clusters: int
            number of clusters for type_clust='kmeans' or type_clust='gmm'
            clustering (could be 0: no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot);
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax') or robust scaling ('robust') or  max absolute scaling ('maxabs')

        col_sample: float
            percentage of features randomly chosen for training

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping

        seed: int
            reproducibility seed for nodes_sim=='uniform', clustering and dropout

        backend: str
            "cpu" or "gpu" or "tpu"

    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu",
    ):
        # input checks -----

        sys_platform = platform.system()

        if (sys_platform == "Windows") and (backend in ("gpu", "tpu")):
            warnings.warn(
                "No GPU/TPU computing on Windows yet, backend set to 'cpu'"
            )
            backend = "cpu"

        assert activation_name in (
            "relu",
            "tanh",
            "sigmoid",
            "prelu",
            "elu",
        ), "'activation_name' must be in ('relu', 'tanh', 'sigmoid','prelu', 'elu')"

        assert nodes_sim in (
            "sobol",
            "hammersley",
            "uniform",
            "halton",
        ), "'nodes_sim' must be in ('sobol', 'hammersley', 'uniform', 'halton')"

        assert type_clust in (
            "kmeans",
            "gmm",
        ), "'type_clust' must be in ('kmeans', 'gmm')"

        assert (len(type_scaling) == 3) & all(
            type_scaling[i] in ("minmax", "std", "robust", "maxabs")
            for i in range(len(type_scaling))
        ), "'type_scaling' must have length 3, and available scaling methods are 'minmax' scaling, standardization ('std'), robust scaling ('robust') and max absolute ('maxabs')"

        assert (col_sample >= 0) & (
            col_sample <= 1
        ), "'col_sample' must be comprised between 0 and 1 (both included)"

        assert backend in (
            "cpu",
            "gpu",
            "tpu",
        ), "must have 'backend' in ('cpu', 'gpu', 'tpu')"

        self.n_hidden_features = n_hidden_features
        self.activation_name = activation_name
        self.a = a
        self.nodes_sim = nodes_sim
        self.bias = bias
        self.seed = seed
        self.backend = backend
        self.dropout = dropout
        self.direct_link = direct_link
        self.cluster_encode = cluster_encode
        self.type_clust = type_clust
        self.type_scaling = type_scaling
        self.col_sample = col_sample
        self.row_sample = row_sample
        self.n_clusters = n_clusters
        if isinstance(self, RegressorMixin):
            self.type_fit = "regression"
        elif isinstance(self, ClassifierMixin):
            self.type_fit = "classification"
        self.subsampler_ = None
        self.index_col_ = None
        self.index_row_ = True
        self.clustering_obj_ = None
        self.clustering_scaler_ = None
        self.nn_scaler_ = None
        self.scaler_ = None
        self.encoder_ = None
        self.W_ = None
        self.X_ = None
        self.y_ = None
        self.y_mean_ = None
        self.beta_ = None

        # activation function -----
        if sys_platform in ("Linux", "Darwin"):
            activation_options = {
                "relu": ac.relu if (self.backend == "cpu") else jnn.relu,
                "tanh": np.tanh if (self.backend == "cpu") else jnp.tanh,
                "sigmoid": (
                    ac.sigmoid if (self.backend == "cpu") else jnn.sigmoid
                ),
                "prelu": partial(ac.prelu, a=a),
                "elu": (
                    partial(ac.elu, a=a)
                    if (self.backend == "cpu")
                    else partial(jnn.elu, a=a)
                ),
            }
        else:  # on Windows currently, no JAX
            activation_options = {
                "relu": (
                    ac.relu if (self.backend == "cpu") else NotImplementedError
                ),
                "tanh": (
                    np.tanh if (self.backend == "cpu") else NotImplementedError
                ),
                "sigmoid": (
                    ac.sigmoid
                    if (self.backend == "cpu")
                    else NotImplementedError
                ),
                "prelu": partial(ac.prelu, a=a),
                "elu": (
                    partial(ac.elu, a=a)
                    if (self.backend == "cpu")
                    else NotImplementedError
                ),
            }
        self.activation_func = activation_options[activation_name]

    # "preprocessing" methods to be inherited -----

    def encode_clusters(self, X=None, predict=False, scaler=None, **kwargs):  #
        """Create new covariates with kmeans or GMM clustering

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            predict: boolean
                is False on training set and True on test set

            scaler: {object} of class StandardScaler, MinMaxScaler, RobustScaler or MaxAbsScaler
                if scaler has already been fitted on training data (online training), it can be passed here

            **kwargs:
                additional parameters to be passed to the
                clustering method

        Returns:

            Clusters' matrix, one-hot encoded: {array-like}

        """

        np.random.seed(self.seed)

        if X is None:
            X = self.X_

        if isinstance(X, pd.DataFrame):
            X = copy.deepcopy(X.values.astype(float))

        if len(X.shape) == 1:
            X = X.reshape(1, -1)

        if predict is False:  # encode training set
            # scale input data before clustering
            self.clustering_scaler_, scaled_X = mo.scale_covariates(
                X, choice=self.type_scaling[2], scaler=self.clustering_scaler_
            )

            self.clustering_obj_, X_clustered = mo.cluster_covariates(
                scaled_X,
                self.n_clusters,
                self.seed,
                type_clust=self.type_clust,
                **kwargs
            )

            if self.cluster_encode:
                return mo.one_hot_encode(X_clustered, self.n_clusters).astype(
                    np.float16
                )

            return X_clustered.astype(np.float16)

        # if predict == True, encode test set
        X_clustered = self.clustering_obj_.predict(
            self.clustering_scaler_.transform(X)
        )

        if self.cluster_encode == True:
            return mo.one_hot_encode(X_clustered, self.n_clusters).astype(
                np.float16
            )

        return X_clustered.astype(np.float16)

    def create_layer(self, scaled_X, W=None):
        """Create hidden layer.

        Parameters:

            scaled_X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features

            W: {array-like}, shape = [n_features, hidden_features]
                if provided, constructs the hidden layer with W; otherwise computed internally

        Returns:

            Hidden layer matrix: {array-like}

        """

        n_features = scaled_X.shape[1]

        # hash_sim = {
        #         "sobol": generate_sobol,
        #         "hammersley": generate_hammersley,
        #         "uniform": generate_uniform,
        #         "halton": generate_halton
        #     }

        if self.bias is False:  # no bias term in the hidden layer
            if W is None:
                if self.nodes_sim == "sobol":
                    self.W_ = generate_sobol(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                        seed=self.seed,
                    )
                elif self.nodes_sim == "hammersley":
                    self.W_ = generate_hammersley(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                        seed=self.seed,
                    )
                elif self.nodes_sim == "uniform":
                    self.W_ = generate_uniform(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                        seed=self.seed,
                    )
                else:
                    self.W_ = generate_halton(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                        seed=self.seed,
                    )

                assert (
                    scaled_X.shape[1] == self.W_.shape[0]
                ), "check dimensions of covariates X and matrix W"

                return mo.dropout(
                    x=self.activation_func(
                        mo.safe_sparse_dot(
                            a=scaled_X, b=self.W_, backend=self.backend
                        )
                    ),
                    drop_prob=self.dropout,
                    seed=self.seed,
                )

            # W is not none
            assert (
                scaled_X.shape[1] == W.shape[0]
            ), "check dimensions of covariates X and matrix W"

            # self.W_ = W
            return mo.dropout(
                x=self.activation_func(
                    mo.safe_sparse_dot(a=scaled_X, b=W, backend=self.backend)
                ),
                drop_prob=self.dropout,
                seed=self.seed,
            )

        # with bias term in the hidden layer
        if W is None:
            n_features_1 = n_features + 1

            if self.nodes_sim == "sobol":
                self.W_ = generate_sobol(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )
            elif self.nodes_sim == "hammersley":
                self.W_ = generate_hammersley(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )
            elif self.nodes_sim == "uniform":
                self.W_ = generate_uniform(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )
            else:
                self.W_ = generate_halton(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )

            # self.W_ = hash_sim[self.nodes_sim](
            #         n_dims=n_features_1,
            #         n_points=self.n_hidden_features,
            #         seed=self.seed,
            #     )

            return mo.dropout(
                x=self.activation_func(
                    mo.safe_sparse_dot(
                        a=mo.cbind(
                            np.ones(scaled_X.shape[0]),
                            scaled_X,
                            backend=self.backend,
                        ),
                        b=self.W_,
                        backend=self.backend,
                    )
                ),
                drop_prob=self.dropout,
                seed=self.seed,
            )

        # W is not None
        # self.W_ = W
        return mo.dropout(
            x=self.activation_func(
                mo.safe_sparse_dot(
                    a=mo.cbind(
                        np.ones(scaled_X.shape[0]),
                        scaled_X,
                        backend=self.backend,
                    ),
                    b=W,
                    backend=self.backend,
                )
            ),
            drop_prob=self.dropout,
            seed=self.seed,
        )

    def _jax_create_layer(
        self, scaled_X: jnp.ndarray, W: Optional[jnp.ndarray] = None
    ) -> jnp.ndarray:
        """JAX-compatible version of create_layer that exactly matches the original functionality."""
        key = jax.random.PRNGKey(self.seed)
        n_features = scaled_X.shape[1]

        # Generate weights if not provided
        if W is None:
            if self.bias:
                n_features_1 = n_features + 1
                shape = (n_features_1, self.n_hidden_features)
            else:
                shape = (n_features, self.n_hidden_features)

            # JAX-compatible weight generation matching original behavior
            if self.nodes_sim == "sobol":
                W_np = generate_sobol(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )
                W = jnp.asarray(W_np)
            elif self.nodes_sim == "hammersley":
                W_np = generate_hammersley(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )
                W = jnp.asarray(W_np)
            elif self.nodes_sim == "uniform":
                key, subkey = jax.random.split(key)
                W = jax.random.uniform(
                    subkey, shape=shape, minval=-1.0, maxval=1.0
                )
            else:  # halton
                W_np = generate_halton(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )
                W = jnp.asarray(W_np)

            self.W_ = np.array(W)  # Store as numpy for original methods

        # Prepare input with bias if needed
        if self.bias:
            X_with_bias = jnp.hstack(
                [jnp.ones((scaled_X.shape[0], 1)), scaled_X]
            )
            print("X_with_bias shape:", X_with_bias.shape)
            print("W shape:", W.shape)
            linear_output = jnp.dot(X_with_bias, W)
        else:
            linear_output = jnp.dot(scaled_X, W)

        # Apply activation function
        if self.activation_name == "relu":
            activated = jax.nn.relu(linear_output)
        elif self.activation_name == "tanh":
            activated = jnp.tanh(linear_output)
        elif self.activation_name == "sigmoid":
            activated = jax.nn.sigmoid(linear_output)
        else:  # leaky relu
            activated = jax.nn.leaky_relu(linear_output, negative_slope=self.a)

        # Apply dropout
        if self.dropout > 0:
            key, subkey = jax.random.split(key)
            mask = jax.random.bernoulli(
                subkey, p=1 - self.dropout, shape=activated.shape
            )
            activated = jnp.where(mask, activated / (1 - self.dropout), 0)

        return activated

    def cook_training_set(self, y=None, X=None, W=None, **kwargs):
        """Create new hidden features for training set, with hidden layer, center the response.

        Parameters:

            y: array-like, shape = [n_samples]
                Target values

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features

            W: {array-like}, shape = [n_features, hidden_features]
                if provided, constructs the hidden layer via W

        Returns:

            (centered response, direct link + hidden layer matrix): {tuple}

        """

        # either X and y are stored or not
        # assert ((y is None) & (X is None)) | ((y is not None) & (X is not None))
        if self.n_hidden_features > 0:  # has a hidden layer
            assert (
                len(self.type_scaling) >= 2
            ), "must have len(self.type_scaling) >= 2 when self.n_hidden_features > 0"

        if X is None:
            if self.col_sample == 1:
                input_X = self.X_
            else:
                n_features = self.X_.shape[1]
                new_n_features = int(np.ceil(n_features * self.col_sample))
                assert (
                    new_n_features >= 1
                ), "check class attribute 'col_sample' and the number of covariates provided for X"
                np.random.seed(self.seed)
                index_col = np.random.choice(
                    range(n_features), size=new_n_features, replace=False
                )
                self.index_col_ = index_col
                input_X = self.X_[:, self.index_col_]

        else:  # X is not None # keep X vs self.X_
            if isinstance(X, pd.DataFrame):
                X = copy.deepcopy(X.values.astype(float))

            if self.col_sample == 1:
                input_X = X
            else:
                n_features = X.shape[1]
                new_n_features = int(np.ceil(n_features * self.col_sample))
                assert (
                    new_n_features >= 1
                ), "check class attribute 'col_sample' and the number of covariates provided for X"
                np.random.seed(self.seed)
                index_col = np.random.choice(
                    range(n_features), size=new_n_features, replace=False
                )
                self.index_col_ = index_col
                input_X = X[:, self.index_col_]

        if self.n_clusters <= 0:
            # data without any clustering: self.n_clusters is None -----

            if self.n_hidden_features > 0:  # with hidden layer
                self.nn_scaler_, scaled_X = mo.scale_covariates(
                    input_X, choice=self.type_scaling[1], scaler=self.nn_scaler_
                )
                Phi_X = (
                    self.create_layer(scaled_X)
                    if W is None
                    else self.create_layer(scaled_X, W=W)
                )
                Z = (
                    mo.cbind(input_X, Phi_X, backend=self.backend)
                    if self.direct_link is True
                    else Phi_X
                )
                self.scaler_, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0], scaler=self.scaler_
                )
            else:  # no hidden layer
                Z = input_X
                self.scaler_, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0], scaler=self.scaler_
                )

        else:
            # data with clustering: self.n_clusters is not None ----- # keep

            augmented_X = mo.cbind(
                input_X,
                self.encode_clusters(input_X, **kwargs),
                backend=self.backend,
            )

            if self.n_hidden_features > 0:  # with hidden layer
                self.nn_scaler_, scaled_X = mo.scale_covariates(
                    augmented_X,
                    choice=self.type_scaling[1],
                    scaler=self.nn_scaler_,
                )
                Phi_X = (
                    self.create_layer(scaled_X)
                    if W is None
                    else self.create_layer(scaled_X, W=W)
                )
                Z = (
                    mo.cbind(augmented_X, Phi_X, backend=self.backend)
                    if self.direct_link is True
                    else Phi_X
                )
                self.scaler_, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0], scaler=self.scaler_
                )
            else:  # no hidden layer
                Z = augmented_X
                self.scaler_, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0], scaler=self.scaler_
                )

        # Returning model inputs -----
        if mx.is_factor(y) is False:  # regression
            # center y
            if y is None:
                self.y_mean_, centered_y = mo.center_response(self.y_)
            else:
                self.y_mean_, centered_y = mo.center_response(y)

            # y is subsampled
            if self.row_sample < 1:
                n, p = Z.shape

                self.subsampler_ = (
                    SubSampler(
                        y=self.y_, row_sample=self.row_sample, seed=self.seed
                    )
                    if y is None
                    else SubSampler(
                        y=y, row_sample=self.row_sample, seed=self.seed
                    )
                )

                self.index_row_ = self.subsampler_.subsample()

                n_row_sample = len(self.index_row_)
                # regression
                return (
                    centered_y[self.index_row_].reshape(n_row_sample),
                    self.scaler_.transform(
                        Z[self.index_row_, :].reshape(n_row_sample, p)
                    ),
                )
            # y is not subsampled
            # regression
            return (centered_y, self.scaler_.transform(Z))

        # classification
        # y is subsampled
        if self.row_sample < 1:
            n, p = Z.shape

            self.subsampler_ = (
                SubSampler(
                    y=self.y_, row_sample=self.row_sample, seed=self.seed
                )
                if y is None
                else SubSampler(y=y, row_sample=self.row_sample, seed=self.seed)
            )

            self.index_row_ = self.subsampler_.subsample()

            n_row_sample = len(self.index_row_)
            # classification
            return (
                y[self.index_row_].reshape(n_row_sample),
                self.scaler_.transform(
                    Z[self.index_row_, :].reshape(n_row_sample, p)
                ),
            )
        # y is not subsampled
        # classification
        return (y, self.scaler_.transform(Z))

    def cook_test_set(self, X, **kwargs):
        """Transform data from test set, with hidden layer.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features

            **kwargs: additional parameters to be passed to self.encode_cluster

        Returns:

            Transformed test set : {array-like}
        """

        if isinstance(X, pd.DataFrame):
            X = copy.deepcopy(X.values.astype(float))

        if len(X.shape) == 1:
            X = X.reshape(1, -1)

        if (
            self.n_clusters == 0
        ):  # data without clustering: self.n_clusters is None -----
            if self.n_hidden_features > 0:
                # if hidden layer
                scaled_X = (
                    self.nn_scaler_.transform(X)
                    if (self.col_sample == 1)
                    else self.nn_scaler_.transform(X[:, self.index_col_])
                )
                Phi_X = self.create_layer(scaled_X, self.W_)
                if self.direct_link:
                    return self.scaler_.transform(
                        mo.cbind(scaled_X, Phi_X, backend=self.backend)
                    )
                # when self.direct_link == False
                return self.scaler_.transform(Phi_X)
            # if no hidden layer # self.n_hidden_features == 0
            return self.scaler_.transform(X)

        # data with clustering: self.n_clusters > 0 -----
        if self.col_sample == 1:
            predicted_clusters = self.encode_clusters(
                X=X, predict=True, **kwargs
            )
            augmented_X = mo.cbind(X, predicted_clusters, backend=self.backend)
        else:
            predicted_clusters = self.encode_clusters(
                X=X[:, self.index_col_], predict=True, **kwargs
            )
            augmented_X = mo.cbind(
                X[:, self.index_col_], predicted_clusters, backend=self.backend
            )

        if self.n_hidden_features > 0:  # if hidden layer
            scaled_X = self.nn_scaler_.transform(augmented_X)
            Phi_X = self.create_layer(scaled_X, self.W_)
            if self.direct_link:
                return self.scaler_.transform(
                    mo.cbind(augmented_X, Phi_X, backend=self.backend)
                )
            return self.scaler_.transform(Phi_X)

        # if no hidden layer
        return self.scaler_.transform(augmented_X)

    def cook_training_set_jax(self, y=None, X=None, W=None, **kwargs):
        """JAX-compatible version of cook_training_set that maintains side effects."""
        # Initialize random key
        key = jax.random.PRNGKey(self.seed)

        # Convert inputs to JAX arrays
        X = jnp.asarray(X) if X is not None else jnp.asarray(self.X_)
        y = jnp.asarray(y) if y is not None else jnp.asarray(self.y_)

        # Handle column sampling
        if self.col_sample < 1:
            n_features = X.shape[1]
            new_n_features = int(jnp.ceil(n_features * self.col_sample))
            assert new_n_features >= 1, "Invalid col_sample"

            key, subkey = jax.random.split(key)
            index_col = jax.random.choice(
                subkey, n_features, shape=(new_n_features,), replace=False
            )
            self.index_col_ = np.array(
                index_col
            )  # Store as numpy for original methods
            input_X = X[:, index_col]
            n_features = (
                new_n_features  # Update n_features after column sampling
            )
        else:
            input_X = X
            n_features = X.shape[1]

        augmented_X = input_X

        # JAX-compatible scaling
        def jax_scale(data, mean=None, std=None):
            if mean is None:
                mean = jnp.mean(data, axis=0)
            if std is None:
                std = jnp.std(data, axis=0)
            return (data - mean) / (std + 1e-10), mean, std

        # Hidden layer processing
        if self.n_hidden_features > 0:
            # Initialize weights if not provided
            if W is None:
                shape = (n_features, self.n_hidden_features)

                # JAX-compatible weight generation
                if self.nodes_sim == "uniform":
                    key, subkey = jax.random.split(key)
                    W = jax.random.uniform(
                        subkey, shape=shape, minval=-1.0, maxval=1.0
                    ) * (1 / jnp.sqrt(n_features))
                else:
                    # For other sequences, use numpy generation then convert to JAX
                    if self.nodes_sim == "sobol":
                        W_np = generate_sobol(
                            n_dims=shape[0],
                            n_points=shape[1],
                            seed=self.seed,
                        )
                    elif self.nodes_sim == "hammersley":
                        W_np = generate_hammersley(
                            n_dims=shape[0],
                            n_points=shape[1],
                            seed=self.seed,
                        )
                    elif self.nodes_sim == "halton":
                        W_np = generate_halton(
                            n_dims=shape[0],
                            n_points=shape[1],
                            seed=self.seed,
                        )
                    else:  # default to uniform
                        key, subkey = jax.random.split(key)
                        W = jax.random.uniform(
                            subkey, shape=shape, minval=-1.0, maxval=1.0
                        ) * (1 / jnp.sqrt(n_features))

                    if self.nodes_sim in ["sobol", "hammersley", "halton"]:
                        W = jnp.asarray(W_np) * (1 / jnp.sqrt(n_features))

                self.W_ = np.array(W)  # Store as numpy for original methods

            # Scale features
            scaled_X, self.nn_mean_, self.nn_std_ = jax_scale(
                augmented_X,
                getattr(self, "nn_mean_", None),
                getattr(self, "nn_std_", None),
            )

            # Create hidden layer with proper bias handling
            linear_output = jnp.dot(scaled_X, W)

            # Apply activation
            if self.activation_name == "relu":
                Phi_X = jax.nn.relu(linear_output)
            elif self.activation_name == "tanh":
                Phi_X = jnp.tanh(linear_output)
            elif self.activation_name == "sigmoid":
                Phi_X = jax.nn.sigmoid(linear_output)
            else:  # leaky relu
                Phi_X = jax.nn.leaky_relu(linear_output, negative_slope=self.a)

            # Apply dropout
            if self.dropout > 0:
                key, subkey = jax.random.split(key)
                mask = jax.random.bernoulli(
                    subkey, p=1 - self.dropout, shape=Phi_X.shape
                )
                Phi_X = jnp.where(mask, Phi_X / (1 - self.dropout), 0)

            Z = jnp.hstack([scaled_X, Phi_X]) if self.direct_link else Phi_X
        else:
            Z = augmented_X

        # Final scaling
        scaled_Z, self.scale_mean_, self.scale_std_ = jax_scale(
            Z,
            getattr(self, "scale_mean_", None),
            getattr(self, "scale_std_", None),
        )

        # Center response for regression
        if not hasattr(mx, "is_factor") or not mx.is_factor(
            y
        ):  # regression case
            self.y_mean_ = float(
                jnp.mean(y)
            )  # Convert to Python float for compatibility
            centered_y = y - self.y_mean_
        else:
            centered_y = y

        # Handle row sampling
        if self.row_sample < 1:
            key, subkey = jax.random.split(key)
            n_samples = Z.shape[0]
            n_row_sample = int(jnp.ceil(n_samples * self.row_sample))
            index_row = jax.random.choice(
                subkey, n_samples, shape=(n_row_sample,), replace=False
            )
            self.index_row_ = np.array(
                index_row
            )  # Store as numpy for original methods
            return (centered_y[index_row], scaled_Z[index_row])

        return (centered_y, scaled_Z)

    def cook_test_set_jax(self, X, **kwargs):
        """JAX-compatible test set processing with matching dimension handling."""
        X = jnp.asarray(X)

        if len(X.shape) == 1:
            X = X.reshape(1, -1)

        # Handle column sampling
        input_X = (
            X if self.col_sample == 1 else X[:, jnp.asarray(self.index_col_)]
        )

        augmented_X = input_X

        # JAX-compatible scaling
        scaled_X = (augmented_X - self.nn_mean_) / (self.nn_std_ + 1e-10)

        # Process hidden layer if needed
        if self.n_hidden_features > 0:
            Phi_X = self._jax_create_layer(scaled_X, jnp.asarray(self.W_))
            Z = jnp.hstack([scaled_X, Phi_X]) if self.direct_link else Phi_X
        else:
            Z = augmented_X

        # Final scaling
        scaled_Z = (Z - self.scale_mean_) / (self.scale_std_ + 1e-10)

        return scaled_Z

    def _jax_create_layer(self, X, W):
        """JAX-compatible hidden layer creation."""
        # print("X", X.shape)
        # print("W", W.shape)
        # print("self.W_", self.W_.shape)
        linear_output = jnp.dot(X, W)

        if self.activation_name == "relu":
            return jax.nn.relu(linear_output)
        elif self.activation_name == "tanh":
            return jnp.tanh(linear_output)
        elif self.activation_name == "sigmoid":
            return jax.nn.sigmoid(linear_output)
        else:  # leaky relu
            return jax.nn.leaky_relu(linear_output, negative_slope=self.a)

    def cross_val_score(
        self,
        X,
        y,
        cv=5,
        scoring="accuracy",
        random_state=42,
        n_jobs=-1,
        epsilon=0.5,
        penalized=True,
        objective="abs",
        **kwargs
    ):
        """
        Penalized Cross-validation score for a model.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features

            y: array-like, shape = [n_samples]
                Target values

            X_test: {array-like}, shape = [n_samples, n_features]
                Test vectors, where n_samples is the number
                of samples and n_features is the number of features

            y_test: array-like, shape = [n_samples]
                Target values

            cv: int
                Number of folds

            scoring: str
                Scoring metric

            random_state: int
                Random state

            n_jobs: int
                Number of jobs to run in parallel

            epsilon: float
                Penalty parameter

            penalized: bool
                Whether to obtain penalized cross-validation score or not

            objective: str
                'abs': Minimize the absolute difference between cross-validation score and validation score
                'relative': Minimize the relative difference between cross-validation score and validation score
        Returns:

            A namedtuple with the following fields:
                - cv_score: float
                    cross-validation score
                - val_score: float
                    validation score
                - penalized_score: float
                    penalized cross-validation score: cv_score / val_score + epsilon*(1/val_score + 1/cv_score)
                    If higher scoring metric is better, minimize the function result.
                    If lower scoring metric is better, maximize the function result.
        """
        if scoring == "accuracy":
            scoring_func = accuracy_score
        elif scoring == "balanced_accuracy":
            scoring_func = balanced_accuracy_score
        elif scoring == "f1":
            scoring_func = f1_score
        elif scoring == "roc_auc":
            scoring_func = roc_auc_score
        elif scoring == "r2":
            scoring_func = r2_score
        elif scoring == "mse":
            scoring_func = mean_squared_error
        elif scoring == "mae":
            scoring_func = mean_absolute_error
        elif scoring == "mape":
            scoring_func = mean_absolute_percentage_error
        elif scoring == "rmse":

            def scoring_func(y_true, y_pred):
                return np.sqrt(mean_squared_error(y_true, y_pred))

        X_train, X_val, y_train, y_val = train_test_split(
            X, y, test_size=0.2, random_state=random_state
        )

        res = cross_val_score(
            self, X_train, y_train, cv=cv, scoring=scoring, n_jobs=n_jobs
        )  # cross-validation error

        if penalized == False:
            return res

        DescribeResult = namedtuple(
            "DescribeResult", ["cv_score", "val_score", "penalized_score"]
        )

        numerator = res.mean()

        # Evaluate on the (cv+1)-th fold
        preds_val = self.fit(X_train, y_train).predict(X_val)
        try:
            denominator = scoring(y_val, preds_val)  # validation error
        except Exception as e:
            denominator = scoring_func(y_val, preds_val)

        # if higher is better
        if objective == "abs":
            penalized_score = np.abs(numerator - denominator) + epsilon * (
                1 / denominator + 1 / numerator
            )
        elif objective == "relative":
            ratio = numerator / denominator
            penalized_score = np.abs(ratio - 1) + epsilon * (
                1 / denominator + 1 / numerator
            )

        return DescribeResult(
            cv_score=numerator,
            val_score=denominator,
            penalized_score=penalized_score,
        )


class RidgeRegressor(Base, RegressorMixin):
    """Ridge Regression model with JAX support.

    Parameters:
        lambda_: float or array-like
            Ridge regularization parameter(s). Default is 0.

        All other parameters are inherited from Base class.
    """

    def __init__(
        self,
        lambda_=0.0,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend,
        )
        self.lambda_ = lambda_

    def _center_scale_xy(self, X, y):
        """Center X and y, scale X."""
        n = X.shape[0]

        # Center X and y
        X_mean = jnp.mean(X, axis=0)
        y_mean = jnp.mean(y)
        X_centered = X - X_mean
        y_centered = y - y_mean

        # Scale X
        X_scale = jnp.sqrt(jnp.sum(X_centered**2, axis=0) / n)
        X_scaled = X_centered / X_scale

        return X_scaled, y_centered, X_mean, y_mean, X_scale

    def fit(self, X, y, **kwargs):
        """Fit Ridge regression model.

        Parameters:
            X : array-like of shape (n_samples, n_features)
                Training data
            y : array-like of shape (n_samples,)
                Target values

        Returns:
            self : returns an instance of self.
        """
        X, y = self.cook_training_set(y, X)

        if self.backend == "cpu":
            # Use numpy for CPU
            (
                X_scaled,
                y_centered,
                self.X_mean_,
                self.y_mean_,
                self.X_scale_,
            ) = self._center_scale_xy(np.array(X), np.array(y))

            # SVD decomposition
            U, d, Vt = np.linalg.svd(X_scaled, full_matrices=False)

            # Compute coefficients
            rhs = U.T @ y_centered
            d2 = d**2

            if np.isscalar(self.lambda_):
                div = d2 + self.lambda_
                a = (d * rhs) / div
                self.coef_ = (Vt.T @ a) / self.X_scale_
            else:
                coefs = []
                for lambda_ in self.lambda_:
                    div = d2 + lambda_
                    a = (d * rhs) / div
                    coef = (Vt.T @ a) / self.X_scale_
                    coefs.append(coef)
                self.coef_ = np.array(coefs).T

        else:
            # Use JAX for GPU/TPU
            (
                X_scaled,
                y_centered,
                self.X_mean_,
                self.y_mean_,
                self.X_scale_,
            ) = self._center_scale_xy(jnp.array(X), jnp.array(y))

            # SVD decomposition
            U, d, Vt = jnp.linalg.svd(X_scaled, full_matrices=False)

            # Compute coefficients
            rhs = mo.safe_sparse_dot(U.T, y_centered, backend=self.backend)
            d2 = d**2

            if np.isscalar(self.lambda_):
                div = d2 + self.lambda_
                a = (d * rhs) / div
                self.coef_ = (
                    mo.safe_sparse_dot(Vt.T, a, backend=self.backend)
                    / self.X_scale_
                )
            else:
                coefs = []
                for lambda_ in self.lambda_:
                    div = d2 + lambda_
                    a = (d * rhs) / div
                    coef = (
                        mo.safe_sparse_dot(Vt.T, a, backend=self.backend)
                        / self.X_scale_
                    )
                    coefs.append(coef)
                self.coef_ = jnp.array(coefs).T

        # Compute GCV, HKB and LW criteria
        y_pred = self.predict(X)
        resid = y - y_pred
        n, p = X.shape
        s2 = jnp.sum(resid**2) / (n - p)

        if self.backend == "cpu":
            self.HKB_ = (p - 2) * s2 / jnp.sum(self.coef_**2)
            self.LW_ = (p - 2) * s2 * n / jnp.sum(y_pred**2)

            if np.isscalar(self.lambda_):
                div = d2 + self.lambda_
                self.GCV_ = (
                    jnp.sum((y - y_pred) ** 2) / (n - jnp.sum(d2 / div)) ** 2
                )
            else:
                self.GCV_ = []
                for lambda_ in self.lambda_:
                    div = d2 + lambda_
                    gcv = (
                        jnp.sum((y - y_pred) ** 2)
                        / (n - jnp.sum(d2 / div)) ** 2
                    )
                    self.GCV_.append(gcv)
                self.GCV_ = jnp.array(self.GCV_)
        else:
            self.HKB_ = (p - 2) * s2 / jnp.sum(self.coef_**2)
            self.LW_ = (p - 2) * s2 * n / jnp.sum(y_pred**2)

            if np.isscalar(self.lambda_):
                div = d2 + self.lambda_
                self.GCV_ = (
                    jnp.sum((y - y_pred) ** 2) / (n - jnp.sum(d2 / div)) ** 2
                )
            else:
                gcvs = []
                for lambda_ in self.lambda_:
                    div = d2 + lambda_
                    gcv = (
                        jnp.sum((y - y_pred) ** 2)
                        / (n - jnp.sum(d2 / div)) ** 2
                    )
                    gcvs.append(gcv)
                self.GCV_ = jnp.array(gcvs)

        return self

    def predict(self, X):
        """Predict using the Ridge regression model.

        Parameters:
            X : array-like of shape (n_samples, n_features)
                Samples to predict for

        Returns:
            y_pred : array-like of shape (n_samples,)
                Returns predicted values.
        """
        X = self.cook_test_set(X)

        if self.backend == "cpu":
            if np.isscalar(self.lambda_):
                return (
                    mo.safe_sparse_dot(X, self.coef_, backend=self.backend)
                    + self.y_mean_
                )
            else:
                return jnp.array(
                    [
                        mo.safe_sparse_dot(X, coef, backend=self.backend)
                        + self.y_mean_
                        for coef in self.coef_.T
                    ]
                ).T
        else:
            if np.isscalar(self.lambda_):
                return (
                    mo.safe_sparse_dot(X, self.coef_, backend=self.backend)
                    + self.y_mean_
                )
            else:
                return jnp.array(
                    [
                        mo.safe_sparse_dot(X, coef, backend=self.backend)
                        + self.y_mean_
                        for coef in self.coef_.T
                    ]
                ).T



import jax
import jax.numpy as jnp
import nnetsauce as ns  # adjust if your import path differs
import pandas as pd
import numpy as np

from .custom import Custom
from sklearn.base import RegressorMixin
from copy import deepcopy
from collections import namedtuple
from tqdm import tqdm
from sklearn.base import BaseEstimator, RegressorMixin
from scipy.stats import norm
from sklearn.metrics import mean_pinball_loss

# import your matrix operations helper if needed (mo.rbind)


class CustomBackPropRegressor(Custom, RegressorMixin):
    """
    Finite difference trainer for nnetsauce models.

    Parameters
    ----------

    base_model : str
        The name of the base model (e.g., 'RidgeCV').

    type_grad : {'finitediff', 'autodiff'}, optional
        Type of gradient computation to use (default='finitediff').

    lr : float, optional
        Learning rate for optimization (default=1e-4).

    optimizer : {'gd', 'sgd', 'adam', 'cd'}, optional
        Optimization algorithm: gradient descent ('gd'), stochastic gradient descent ('sgd'),
        Adam ('adam'), or coordinate descent ('cd'). Default is 'gd'.

    eps : float, optional
        Scaling factor for adaptive finite difference step size (default=1e-3).

    batch_size : int, optional
        Batch size for 'sgd' optimizer (default=32).

    alpha : float, optional
        Elastic net penalty strength (default=0.0).

    l1_ratio : float, optional
        Elastic net mixing parameter (0 = Ridge, 1 = Lasso, default=0.0).

    type_loss : {'mse', 'quantile'}, optional
        Type of loss function to use (default='mse').

    q : float, optional
        Quantile for quantile loss (default=0.5).

    **kwargs
        Additional parameters to pass to the scikit-learn model.

    """

    def __init__(
        self,
        base_model,
        type_grad="finitediff",
        lr=1e-4,
        optimizer="gd",
        eps=1e-3,
        batch_size=32,
        alpha=0.0,
        l1_ratio=0.0,
        type_loss="mse",
        q=0.5,
        backend="cpu",
        **kwargs,
    ):
        super().__init__(base_model, True, **kwargs)
        self.base_model = base_model
        self.custom_kwargs = kwargs
        self.backend = backend
        self.model = ns.CustomRegressor(
            self.base_model, backend=self.backend, **self.custom_kwargs
        )
        assert isinstance(
            self.model, ns.CustomRegressor
        ), "'model' must be of class ns.CustomRegressor"
        self.type_grad = type_grad
        self.lr = lr
        self.optimizer = optimizer
        self.eps = eps
        self.loss_history_ = []
        self.opt_state = None
        self.batch_size = batch_size  # for SGD
        self.loss_history_ = []
        self._cd_index = 0  # For coordinate descent
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.type_loss = type_loss
        self.q = q

    def _loss(self, X, y, **kwargs):
        """
        Compute the loss (with elastic net penalty) for the current model.

        Parameters
        ----------

        X : array-like of shape (n_samples, n_features)
            Input data.

        y : array-like of shape (n_samples,)
            Target values.

        **kwargs
            Additional keyword arguments for loss calculation.

        Returns
        -------
        float
            The computed loss value.
        """
        y_pred = self.model.predict(X)
        if self.type_loss == "mse":
            loss = np.mean((y - y_pred) ** 2)
        elif self.type_loss == "quantile":
            loss = mean_pinball_loss(y, y_pred, alpha=self.q, **kwargs)
        W = self.model.W_
        l1 = np.sum(np.abs(W))
        l2 = np.sum(W**2)
        return loss + self.alpha * (
            self.l1_ratio * l1 + 0.5 * (1 - self.l1_ratio) * l2
        )

    def _compute_grad(self, X, y):
        """
        Compute the gradient of the loss with respect to W_ using finite differences.

        Parameters
        ----------

        X : array-like of shape (n_samples, n_features)
            Input data.

        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------

        ndarray
            Gradient array with the same shape as W_.
        """
        if self.type_grad == "autodiff":
            raise NotImplementedError(
                "Automatic differentiation is not implemented yet."
            )
            # Use JAX for automatic differentiation
            W = deepcopy(self.model.W_)
            W_flat = W.flatten()
            n_params = W_flat.size

            def loss_fn(W_flat):
                W_reshaped = W_flat.reshape(W.shape)
                self.model.W_ = W_reshaped
                return self._loss(X, y)

            grad_fn = jax.grad(loss_fn)
            grad_flat = grad_fn(W_flat)
            grad = grad_flat.reshape(W.shape)

            # Add elastic net gradient
            l1_grad = self.alpha * self.l1_ratio * np.sign(W)
            l2_grad = self.alpha * (1 - self.l1_ratio) * W
            grad += l1_grad + l2_grad

            self.model.W_ = W
            return grad

        # Finite difference gradient computation
        W = deepcopy(self.model.W_)
        shape = W.shape
        W_flat = W.flatten()
        n_params = W_flat.size

        # Adaptive finite difference step
        h_vec = self.eps * np.maximum(1.0, np.abs(W_flat))
        eye = np.eye(n_params)

        loss_plus = np.zeros(n_params)
        loss_minus = np.zeros(n_params)

        for i in range(n_params):
            h_i = h_vec[i]
            Wp = W_flat.copy()
            Wp[i] += h_i
            Wm = W_flat.copy()
            Wm[i] -= h_i

            self.model.W_ = Wp.reshape(shape)
            loss_plus[i] = self._loss(X, y)

            self.model.W_ = Wm.reshape(shape)
            loss_minus[i] = self._loss(X, y)

        grad = ((loss_plus - loss_minus) / (2 * h_vec)).reshape(shape)

        # Add elastic net gradient
        l1_grad = self.alpha * self.l1_ratio * np.sign(W)
        l2_grad = self.alpha * (1 - self.l1_ratio) * W
        grad += l1_grad + l2_grad

        self.model.W_ = W  # restore original
        return grad

    def fit(
        self,
        X,
        y,
        epochs=10,
        verbose=True,
        show_progress=True,
        sample_weight=None,
        **kwargs,
    ):
        """
        Fit the model using finite difference optimization.

        Parameters
        ----------

        X : array-like of shape (n_samples, n_features)
            Training data.

        y : array-like of shape (n_samples,)
            Target values.

        epochs : int, optional
            Number of optimization steps (default=10).

        verbose : bool, optional
            Whether to print progress messages (default=True).

        show_progress : bool, optional
            Whether to show tqdm progress bar (default=True).

        sample_weight : array-like, optional
            Sample weights.

        **kwargs
            Additional keyword arguments.

        Returns
        -------

        self : object
            Returns self.
        """

        self.model.fit(X, y)

        iterator = tqdm(range(epochs)) if show_progress else range(epochs)

        for epoch in iterator:
            grad = self._compute_grad(X, y)

            if self.optimizer == "gd":
                self.model.W_ -= self.lr * grad
                self.model.W_ = np.clip(self.model.W_, 0, 1)
                # print("self.model.W_", self.model.W_)

            elif self.optimizer == "sgd":
                # Sample a mini-batch for stochastic gradient
                n_samples = X.shape[0]
                idxs = np.random.choice(
                    n_samples, self.batch_size, replace=False
                )
                if isinstance(X, pd.DataFrame):
                    X_batch = X.iloc[idxs, :]
                else:
                    X_batch = X[idxs, :]
                y_batch = y[idxs]
                grad = self._compute_grad(X_batch, y_batch)

                self.model.W_ -= self.lr * grad
                self.model.W_ = np.clip(self.model.W_, 0, 1)

            elif self.optimizer == "adam":
                if self.opt_state is None:
                    self.opt_state = {
                        "m": np.zeros_like(grad),
                        "v": np.zeros_like(grad),
                        "t": 0,
                    }
                beta1, beta2, eps = 0.9, 0.999, 1e-8
                self.opt_state["t"] += 1
                self.opt_state["m"] = (
                    beta1 * self.opt_state["m"] + (1 - beta1) * grad
                )
                self.opt_state["v"] = beta2 * self.opt_state["v"] + (
                    1 - beta2
                ) * (grad**2)
                m_hat = self.opt_state["m"] / (1 - beta1 ** self.opt_state["t"])
                v_hat = self.opt_state["v"] / (1 - beta2 ** self.opt_state["t"])

                self.model.W_ -= self.lr * m_hat / (np.sqrt(v_hat) + eps)
                self.model.W_ = np.clip(self.model.W_, 0, 1)
                # print("self.model.W_", self.model.W_)

            elif self.optimizer == "cd":  # coordinate descent
                W_shape = self.model.W_.shape
                W_flat_size = self.model.W_.size
                W_flat = self.model.W_.flatten()
                grad_flat = grad.flatten()

                # Update only one coordinate per epoch (cyclic)
                idx = self._cd_index % W_flat_size
                W_flat[idx] -= self.lr * grad_flat[idx]
                # Clip the updated value
                W_flat[idx] = np.clip(W_flat[idx], 0, 1)

                # Restore W_
                self.model.W_ = W_flat.reshape(W_shape)

                self._cd_index += 1

            else:
                raise ValueError(f"Unsupported optimizer: {self.optimizer}")

            loss = self._loss(X, y)
            self.loss_history_.append(loss)

            if verbose:
                print(f"Epoch {epoch+1}: Loss = {loss:.6f}")

        # if sample_weights, else: (must use self.row_index)
        if sample_weight in kwargs:
            self.model.fit(
                X,
                y,
                sample_weight=sample_weight[self.index_row_].ravel(),
                **kwargs,
            )

            return self

        return self

    def predict(self, X, level=95, method="splitconformal", **kwargs):
        """
        Predict using the trained model.

        Parameters
        ----------

        X : array-like of shape (n_samples, n_features)
            Input data.

        level : int, optional
            Level of confidence for prediction intervals (default=95).

        method : {'splitconformal', 'localconformal'}, optional
            Method for conformal prediction (default='splitconformal').

        **kwargs
            Additional keyword arguments. Use `return_pi=True` for prediction intervals,
            or `return_std=True` for standard deviation estimates.

        Returns
        -------

        array or tuple
            Model predictions, or a tuple with prediction intervals or standard deviations if requested.
        """
        if "return_std" in kwargs:
            alpha = 100 - level
            pi_multiplier = norm.ppf(1 - alpha / 200)

            if len(X.shape) == 1:
                n_features = X.shape[0]
                new_X = mo.rbind(
                    X.reshape(1, n_features),
                    np.ones(n_features).reshape(1, n_features),
                )

                mean_, std_ = self.model.predict(new_X, return_std=True)[0]

                preds = mean_
                lower = mean_ - pi_multiplier * std_
                upper = mean_ + pi_multiplier * std_

                DescribeResults = namedtuple(
                    "DescribeResults", ["mean", "std", "lower", "upper"]
                )

                return DescribeResults(preds, std_, lower, upper)

            # len(X.shape) > 1
            mean_, std_ = self.model.predict(X, return_std=True)

            preds = mean_
            lower = mean_ - pi_multiplier * std_
            upper = mean_ + pi_multiplier * std_

            DescribeResults = namedtuple(
                "DescribeResults", ["mean", "std", "lower", "upper"]
            )

            return DescribeResults(preds, std_, lower, upper)

        if "return_pi" in kwargs:
            assert method in (
                "splitconformal",
                "localconformal",
            ), "method must be in ('splitconformal', 'localconformal')"
            self.pi = ns.PredictionInterval(
                obj=self,
                method=method,
                level=level,
                type_pi=self.type_pi,
                replications=self.replications,
                kernel=self.kernel,
            )

            if len(self.X_.shape) == 1:
                if isinstance(X, pd.DataFrame):
                    self.X_ = pd.DataFrame(
                        self.X_.values.reshape(1, -1), columns=self.X_.columns
                    )
                else:
                    self.X_ = self.X_.reshape(1, -1)
                self.y_ = np.array([self.y_])

            self.pi.fit(self.X_, self.y_)
            # self.X_ = None # consumes memory to keep, dangerous to delete (side effect)
            # self.y_ = None # consumes memory to keep, dangerous to delete (side effect)
            preds = self.pi.predict(X, return_pi=True)
            return preds

        # "return_std" not in kwargs
        if len(X.shape) == 1:
            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (0 + self.model.predict(new_X, **kwargs))[0]

        # len(X.shape) > 1
        return self.model.predict(X, **kwargs)



import copy
import numpy as np
import pandas as pd

from .mts import MTS
from ..utils import matrixops as mo
from ..utils import timeseries as ts
from sklearn.model_selection import GridSearchCV
from sklearn.neighbors import KernelDensity
from tqdm import tqdm

class MultiOutputMTS(MTS):
    """MTS subclass optimized for multivariate time series with vectorized models
    
    Enforces n_series >= 2 and uses single vectorized fit call instead of per-series loop.
    Works with sklearn models supporting multi-output (Ridge, Lasso, LinearRegression, etc.)
    """
    
    def fit(self, X, xreg=None, **kwargs):
        """Fit with vectorized multi-output model - requires n_series >= 2"""
        
        # Validate multivariate input
        try:
            self.init_n_series_ = X.shape[1]
        except IndexError:
            raise ValueError("MultiOutputMTS requires multivariate input (n_samples, n_series)")
        
        if self.init_n_series_ < 2:
            raise ValueError(f"MultiOutputMTS requires at least 2 series, got {self.init_n_series_}")
        
        # Automatic lag selection if requested (copied from parent)
        if isinstance(self.lags, str):
            max_lags = min(25, X.shape[0] // 4)
            best_ic = float("inf")
            best_lags = 1
            
            if self.verbose:
                print(f"\nSelecting optimal number of lags using {self.lags}...")
                iterator = tqdm(range(1, max_lags + 1))
            else:
                iterator = range(1, max_lags + 1)
            
            for lag in iterator:
                if isinstance(X, pd.DataFrame):
                    X_values = X.values[::-1]
                else:
                    X_values = X[::-1]
                
                mts_input = ts.create_train_inputs(X_values, lag)
                dummy_y, scaled_Z = self.cook_training_set(
                    y=np.ones(mts_input[0].shape[0]), X=mts_input[1]
                )
                
                # Vectorized fit for lag selection
                y_means = np.mean(mts_input[0], axis=0)
                centered_y = mts_input[0] - y_means[np.newaxis, :]
                self.obj.fit(X=scaled_Z, y=centered_y)
                residuals = centered_y - self.obj.predict(scaled_Z)
                self.residuals_ = residuals  # Keep (n_obs, n_series) shape
                
                ic = self._compute_information_criterion(curr_lags=lag, criterion=self.lags)
                
                if self.verbose:
                    print(f"Trying lags={lag}, {self.lags}={ic:.2f}")
                
                if ic < best_ic:
                    best_ic = ic
                    best_lags = lag
            
            if self.verbose:
                print(f"\nSelected {best_lags} lags with {self.lags}={best_ic:.2f}")
            
            self.lags = best_lags
        
        # Data preprocessing (from parent)
        self.input_dates = None
        self.df_ = None
        
        if isinstance(X, pd.DataFrame) is False:
            if xreg is None:
                X = pd.DataFrame(X)
                self.series_names = ["series" + str(i) for i in range(X.shape[1])]
            else:
                X = mo.cbind(X, xreg)
                self.xreg_ = xreg
        else:
            X_index = None
            if X.index is not None:
                X_index = X.index
            if xreg is None:
                X = copy.deepcopy(mo.convert_df_to_numeric(X))
            else:
                X = copy.deepcopy(mo.cbind(mo.convert_df_to_numeric(X), xreg))
                self.xreg_ = xreg
            if X_index is not None:
                X.index = X_index
            self.series_names = X.columns.tolist()
        
        if isinstance(X, pd.DataFrame):
            if self.df_ is None:
                self.df_ = X
                X = X.values
            else:
                input_dates_prev = pd.DatetimeIndex(self.df_.index.values)
                frequency = pd.infer_freq(input_dates_prev)
                self.df_ = pd.concat([self.df_, X], axis=0)
                self.input_dates = pd.date_range(
                    start=input_dates_prev[0],
                    periods=len(input_dates_prev) + X.shape[0],
                    freq=frequency,
                ).values.tolist()
                self.df_.index = self.input_dates
                X = self.df_.values
            self.df_.columns = self.series_names
        else:
            if self.df_ is None:
                self.df_ = pd.DataFrame(X, columns=self.series_names)
            else:
                self.df_ = pd.concat(
                    [self.df_, pd.DataFrame(X, columns=self.series_names)],
                    axis=0,
                )
        
        self.input_dates = ts.compute_input_dates(self.df_)
        
        n, p = X.shape
        self.n_obs_ = n
        rep_1_n = np.repeat(1, n)
        
        self.y_ = None
        self.X_ = None
        self.n_series = p
        self.fit_objs_.clear()
        self.y_means_.clear()
        self.residuals_ = None
        self.residuals_sims_ = None
        self.kde_ = None
        self.sims_ = None
        self.scaled_Z_ = None
        self.centered_y_is_ = []
        
        # Create training inputs
        mts_input = ts.create_train_inputs(X[::-1], self.lags)
        self.y_ = mts_input[0]
        self.X_ = mts_input[1]
        
        dummy_y, scaled_Z = self.cook_training_set(y=rep_1_n, X=self.X_)
        self.scaled_Z_ = scaled_Z
        
        if self.verbose > 0:
            print(f"\n Adjusting {type(self.obj).__name__} to multivariate time series (vectorized)... \n")
        
        # VECTORIZED FITTING - NO LOOP
        y_means_array = np.array([np.mean(self.y_[:, i]) for i in range(self.init_n_series_)])
        for i in range(self.init_n_series_):
            self.y_means_[i] = y_means_array[i]
        
        centered_y_all = self.y_ - y_means_array[np.newaxis, :]
        self.centered_y_is_ = [centered_y_all[:, i] for i in range(self.init_n_series_)]
        
        # Single vectorized fit for all series
        self.obj.fit(scaled_Z, centered_y_all)
        
        # All series share the same model
        for i in range(self.init_n_series_):
            self.fit_objs_[i] = self.obj
        
        # Vectorized residuals - ONLY target columns (n_obs, n_series)
        preds_all = self.obj.predict(scaled_Z)
        residuals_raw = centered_y_all - preds_all
        
        # CRITICAL: Ensure residuals only have n_series columns, not all scaled_Z columns
        # In case there's some dimension mismatch, explicitly slice
        self.residuals_ = residuals_raw[:, :self.init_n_series_]
        
        # Handle type_pi
        if self.type_pi == "gaussian":
            self.gaussian_preds_std_ = np.std(self.residuals_, axis=0)
        
        if self.type_pi.startswith("scp2"):
            data_mean = np.mean(self.residuals_, axis=0)
            self.residuals_std_dev_ = np.std(self.residuals_, axis=0)
            self.residuals_ = (self.residuals_ - data_mean[np.newaxis, :]) / self.residuals_std_dev_[np.newaxis, :]
        
        if self.replications is not None and "kde" in self.type_pi:
            if self.verbose > 0:
                print(f"\n Simulate residuals using {self.kernel} kernel... \n")
            assert self.kernel in ("gaussian", "tophat"), "currently, 'kernel' must be either 'gaussian' or 'tophat'"
            kernel_bandwidths = {"bandwidth": np.logspace(-6, 6, 150)}
            grid = GridSearchCV(
                KernelDensity(kernel=self.kernel, **kwargs),
                param_grid=kernel_bandwidths,
            )
            grid.fit(self.residuals_)
            if self.verbose > 0:
                print(f"\n Best parameters for {self.kernel} kernel: {grid.best_params_} \n")
            self.kde_ = grid.best_estimator_
        
        return self
    
    def predict(self, h=5, level=95, quantiles=None, **kwargs):
        """Override predict to handle vectorized model predictions"""
        
        # Delegate to parent for quantiles and multiple levels
        if quantiles is not None or isinstance(level, (list, np.ndarray)):
            return super().predict(h=h, level=level, quantiles=quantiles, **kwargs)
        
        # Store original obj temporarily
        original_obj = self.obj
        
        # Create wrapper that extracts the i-th output for each series
        class VectorizedWrapper:
            def __init__(self, model, series_idx):
                self.model = model
                self.series_idx = series_idx
            
            def predict(self, X, **kw):
                """Predict and return only the output for this series index"""
                preds = self.model.predict(X, **kw)
                # preds shape: (n_samples, n_series) or (n_series,)
                if len(preds.shape) == 1:
                    # Single prediction: (n_series,)
                    return preds[self.series_idx:self.series_idx+1]
                else:
                    # Multiple predictions: (n_samples, n_series)
                    return preds[:, self.series_idx:self.series_idx+1].flatten()
        
        # Wrap each series with its own index
        for i in range(self.init_n_series_):
            self.fit_objs_[i] = VectorizedWrapper(original_obj, i)
        
        try:
            result = super().predict(h=h, level=level, quantiles=quantiles, **kwargs)
        finally:
            # Restore original
            for i in range(self.init_n_series_):
                self.fit_objs_[i] = original_obj
        
        return result


# Authors: T. Moudiki
#
# License: BSD 3 Clause Clear

import numpy as np
import sklearn.metrics as skm2
from sklearn.metrics import mean_pinball_loss
from .glm import GLM
from ..utils import matrixops as mo
from sklearn.base import RegressorMixin
from ..optimizers import Optimizer

try:
    import jax.numpy as jnp
except ImportError:
    pass


class GLMRegressor(GLM, RegressorMixin):
    """Generalized 'linear' models using quasi-randomized networks (regression)

    Attributes:

        n_hidden_features: int
            number of nodes in the hidden layer

        lambda1: float
            regularization parameter for GLM coefficients on original features

        alpha1: float
            controls compromize between l1 and l2 norm of GLM coefficients on original features

        lambda2: float
            regularization parameter for GLM coefficients on nonlinear features

        alpha2: float
            controls compromize between l1 and l2 norm of GLM coefficients on nonlinear features

        family: str
            "gaussian", "laplace", "poisson", or "quantile" (for now)

        level: int, default=50
            The level of the quantiles to compute for family = "quantile".
            Default is the median.

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton',
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model's
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0:
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        optimizer: object
            optimizer, from class nnetsauce.utils.Optimizer

        backend: str.
            "cpu" or "gpu" or "tpu".

        seed: int
            reproducibility seed for nodes_sim=='uniform'

        backend: str
            "cpu", "gpu", "tpu"

    Attributes:

        beta_: vector
            regression coefficients

    Examples:

        See [https://github.com/Techtonique/nnetsauce/blob/master/examples/glm_regression.py](https://github.com/Techtonique/nnetsauce/blob/master/examples/glm_regression.py)

    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        lambda1=0.01,
        alpha1=0.5,
        lambda2=0.01,
        alpha2=0.5,
        family="gaussian",
        level=50,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        optimizer=Optimizer(),
        backend="cpu",
        seed=123,
    ):
        super().__init__(
            n_hidden_features=n_hidden_features,
            lambda1=lambda1,
            alpha1=alpha1,
            lambda2=lambda2,
            alpha2=alpha2,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            optimizer=optimizer,
            backend=backend,
            seed=seed,
        )

        self.family = family
        self.level = level
        self.q = self.level / 100

    def gaussian_loss(self, y, row_index, XB):
        return 0.5 * np.mean(np.square(y[row_index] - XB))

    def laplace_loss(self, y, row_index, XB):
        return 0.5 * np.mean(np.abs(y[row_index] - XB))

    def poisson_loss(self, y, row_index, XB):
        return -np.mean(y[row_index] * XB - np.exp(XB))

    def pinball_loss(self, y, row_index, XB, tau=0.5):
        y = np.array(y[row_index])
        y_pred = np.array(XB)
        return mean_pinball_loss(y, y_pred, alpha=tau)
        # return np.mean(np.maximum(tau * residuals, (tau - 1) * residuals))

    def loss_func(
        self,
        beta,
        group_index,
        X,
        y,
        row_index=None,
        type_loss="gaussian",
        **kwargs
    ):
        res = {
            "gaussian": self.gaussian_loss,
            "laplace": self.laplace_loss,
            "poisson": self.poisson_loss,
            "quantile": self.pinball_loss,
        }

        if type_loss != "quantile":
            if row_index is None:
                row_index = range(len(y))
                XB = self.compute_XB(X, beta=beta)

                return res[type_loss](y, row_index, XB) + self.compute_penalty(
                    group_index=group_index, beta=beta
                )

            XB = self.compute_XB(X, beta=beta, row_index=row_index)

            return res[type_loss](y, row_index, XB) + self.compute_penalty(
                group_index=group_index, beta=beta
            )

        else:  # quantile
            assert (
                self.q > 0 and self.q < 1
            ), "'tau' must be comprised 0 < tau < 1"

            if row_index is None:
                row_index = range(len(y))
                XB = self.compute_XB(X, beta=beta)
                return res[type_loss](y, row_index, XB, self.q)

            XB = self.compute_XB(X, beta=beta, row_index=row_index)
            return res[type_loss](y, row_index, XB, self.q)

    def fit(self, X, y, **kwargs):
        """Fit GLM model to training data (X, y).

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to
                    self.cook_training_set or self.obj.fit

        Returns:

            self: object

        """
        self.beta_ = None
        self.n_iter = 0

        _, self.group_index = X.shape

        centered_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)
        # initialization
        if self.backend == "cpu":
            beta_ = np.linalg.lstsq(scaled_Z, centered_y, rcond=None)[0]
        else:
            beta_ = jnp.linalg.lstsq(scaled_Z, centered_y, rcond=None)[0]
        # optimization
        # fit(self, loss_func, response, x0, **kwargs):
        # loss_func(self, beta, group_index, X, y,
        #          row_index=None, type_loss="gaussian",
        #          **kwargs)
        self.optimizer.fit(
            self.loss_func,
            response=centered_y,
            x0=beta_,
            group_index=self.group_index,
            X=scaled_Z,
            y=centered_y,
            type_loss=self.family,
            **kwargs
        )

        self.beta_ = self.optimizer.results[0]

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            model predictions: {array-like}

        """

        if len(X.shape) == 1:
            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (
                self.y_mean_
                + np.dot(self.cook_test_set(new_X, **kwargs), self.beta_)
            )[0]

        return self.y_mean_ + np.dot(
            self.cook_test_set(X, **kwargs), self.beta_
        )

    def score(self, X, y, scoring=None):
        """Compute the score of the model.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            scoring: str
                scoring method

        Returns:

            score: float

        """

        if scoring is None:
            return np.sqrt(np.mean((self.predict(X) - y) ** 2))

        return skm2.get_scorer(scoring)(self, X, y)



# Authors: Thierry Moudiki
#
# License: BSD 3 Clear

import numpy as np
import platform
from scipy.optimize import minimize
import sklearn.metrics as skm
from .ridge2 import Ridge2
from ..utils import matrixops as mo
from ..utils import misc as mx
from sklearn.base import RegressorMixin
from scipy.special import logsumexp
from scipy.linalg import pinv

try:
    from jax.numpy.linalg import pinv as jpinv
except ImportError:
    pass


class Ridge2Regressor(Ridge2, RegressorMixin):
    """Ridge regression with 2 regularization parameters derived from class Ridge

    Parameters:

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton',
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out
            of the training

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0:
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        lambda1: float
            regularization parameter on direct link

        lambda2: float
            regularization parameter on hidden layer

        seed: int
            reproducibility seed for nodes_sim=='uniform'

        backend: str
            'cpu' or 'gpu' or 'tpu'

    Attributes:

        beta_: {array-like}
            regression coefficients

        coef_: {array-like}
            alias for `beta_`, regression coefficients

        y_mean_: float
            average response

    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        lambda1=0.1,
        lambda2=0.1,
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            lambda1=lambda1,
            lambda2=lambda2,
            seed=seed,
            backend=backend,
        )

        self.type_fit = "regression"
        self.coef_ = None

    def fit(self, X, y, **kwargs):
        """Fit Ridge model to training data (X, y).

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to
                    self.cook_training_set or self.obj.fit

        Returns:

            self: object

        """

        sys_platform = platform.system()

        centered_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)

        n_X, p_X = X.shape
        n_Z, p_Z = scaled_Z.shape

        if self.n_clusters > 0:
            if self.encode_clusters == True:
                n_features = p_X + self.n_clusters
            else:
                n_features = p_X + 1
        else:
            n_features = p_X

        X_ = scaled_Z[:, 0:n_features]
        Phi_X_ = scaled_Z[:, n_features:p_Z]

        B = mo.crossprod(x=X_, backend=self.backend) + self.lambda1 * np.diag(
            np.repeat(1, n_features)
        )
        C = mo.crossprod(x=Phi_X_, y=X_, backend=self.backend)
        D = mo.crossprod(
            x=Phi_X_, backend=self.backend
        ) + self.lambda2 * np.diag(np.repeat(1, Phi_X_.shape[1]))

        if sys_platform in ("Linux", "Darwin"):
            B_inv = pinv(B) if self.backend == "cpu" else jpinv(B)
        else:
            B_inv = pinv(B)

        W = mo.safe_sparse_dot(a=C, b=B_inv, backend=self.backend)
        S_mat = D - mo.tcrossprod(x=W, y=C, backend=self.backend)

        if sys_platform in ("Linux", "Darwin"):
            S_inv = pinv(S_mat) if self.backend == "cpu" else jpinv(S_mat)
        else:
            S_inv = pinv(S_mat)

        Y = mo.safe_sparse_dot(a=S_inv, b=W, backend=self.backend)
        inv = mo.rbind(
            mo.cbind(
                x=B_inv + mo.crossprod(x=W, y=Y, backend=self.backend),
                y=-np.transpose(Y),
                backend=self.backend,
            ),
            mo.cbind(x=-Y, y=S_inv, backend=self.backend),
            backend=self.backend,
        )

        self.beta_ = mo.safe_sparse_dot(
            a=inv,
            b=mo.crossprod(x=scaled_Z, y=centered_y, backend=self.backend),
            backend=self.backend,
        )

        self.coef_ = self.beta_  # sklearn compatibility

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            model predictions: {array-like}

        """

        if len(X.shape) == 1:
            n_features = X.shape[0]
            new_X = mo.rbind(
                x=X.reshape(1, n_features),
                y=np.ones(n_features).reshape(1, n_features),
                backend=self.backend,
            )

            return (
                self.y_mean_
                + mo.safe_sparse_dot(
                    a=self.cook_test_set(new_X, **kwargs),
                    b=self.beta_,
                    backend=self.backend,
                )
            )[0]

        return self.y_mean_ + mo.safe_sparse_dot(
            a=self.cook_test_set(X, **kwargs),
            b=self.beta_,
            backend=self.backend,
        )

    def partial_fit(self, X, y, learning_rate=0.01, decay=0.001, **kwargs):
        """Incrementally fit the Ridge model using SGD-style updates.

        Uses the update rule: w_{n+1} = w_n + γ_n * x_n * [y_n - x_n^T * w_n] - γ_n * λ * w_n
        for online learning with individual samples.

        Args:
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors for this batch

            y: array-like, shape = [n_samples]
                Target values for this batch

            learning_rate: float, default=0.01
                Initial learning rate for SGD updates

            decay: float, default=0.001
                Learning rate decay parameter

            **kwargs: additional parameters to be passed to self.cook_training_set

        Returns:
            self: object
        """

        # Input validation
        X = np.asarray(X)
        y = np.asarray(y)

        if X.shape[0] != y.shape[0]:
            raise ValueError("X and y must have the same number of samples")

        # Handle first call
        if not self._is_fitted:
            # Initialize learning parameters
            self.initial_learning_rate = learning_rate
            self.decay = decay
            self._step_count = 0
            self._is_fitted = True

        # Process the batch
        centered_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)

        # Get dimensions
        n_samples, n_features_total = scaled_Z.shape
        n_original_features = X.shape[1]

        # Determine feature dimensions for regularization
        if self.n_clusters > 0:
            if self.cluster_encode:
                n_direct_features = n_original_features + self.n_clusters
            else:
                n_direct_features = n_original_features + 1
        else:
            n_direct_features = n_original_features

        # Initialize beta_ if first time
        if not hasattr(self, "beta_") or self.beta_ is None:
            # For regression, beta_ is 1D (single output)
            self.beta_ = np.zeros(n_features_total)

        # Precompute indices for regularization
        direct_indices = slice(0, n_direct_features)
        hidden_indices = slice(n_direct_features, n_features_total)

        # Process each sample with SGD
        for i in range(n_samples):
            self._step_count += 1

            # Current learning rate with decay
            current_lr = self.initial_learning_rate / (
                1 + self.decay * self._step_count
            )

            # Current sample and target
            x_i = scaled_Z[i, :]  # Feature vector
            y_i = (
                centered_y[i] if centered_y.ndim == 1 else centered_y[i, 0]
            )  # Scalar target

            # Prediction: x_i^T * beta
            prediction = x_i @ self.beta_

            # Error: y_i - prediction
            error = y_i - prediction

            # Gradient update: current_lr * x_i * error
            gradient_update = current_lr * x_i * error

            # Regularization terms (more efficient indexing)
            reg_update = np.zeros_like(self.beta_)
            reg_update[direct_indices] = (
                current_lr * self.lambda1 * self.beta_[direct_indices]
            )
            reg_update[hidden_indices] = (
                current_lr * self.lambda2 * self.beta_[hidden_indices]
            )

            # Combined update: beta = beta + gradient_update - reg_update
            self.beta_ += gradient_update - reg_update

        self.coef_ = self.beta_  # sklearn compatibility

        return self
